﻿Public Class ACellmodel

    Public R As Double = 8.3143                        'Gas constant
    Public T As Double = 310.15                         'absolute temperature
    Public F As Double = 96.4867                       'Faraday const. Coulomb/mM
    Public RTF As Double = R * T / F                  'for monovalent ions
    Public RTF2 As Double = R * T / F / 2           'for divalent ions

    Public ATPenergy As Double = 48                  'Energy of ATP hydrolysis (kJ/mol) , is not used in the present version.

    Public originalCm As Double = 48             'pF, whole cell membrane capacitance of an Atrial cell

    Public originalGgap As Double = 600      'nS, Ggap between Atrial cells

    Public sumQ As Double = 0             'sum of charge carried by IgapK between SA(Nc) and AC(1)

    'Public ATPenergy As Double = 48                  'Energy of ATP hydrolysis (kJ/mol) , is not used in the present version.

    Const SRaFactor As Double = 0.7
    '***************************(4-A) Volume of compartments***************************
    Public Vi As Double = 115 * 13 * 8 / 2            'um^3  femto Liter, adopted from Wilders  osmoticaly active volume is smaller than the bulk volume. refer to Fujioka et al, 1998
    Public Vrel As Double = 0.02 * Vi * SRaFactor              '0.02 * Vi        Volume of releasing site
    Public Vup As Double = 0.05 * Vi * SRaFactor          '0.05 * Vi        Volume of uptake site

    '*************************** membrane parameters****************************
    Public Cm As Double = 48         'pF, whole cell membrane capacitance
    Public Vm As Double           'mV, membrane potential at time t
    Public dVmdt As Double      'mV/ms, rate of change in Vm
    Public previousVm As Double
    Public latestVm As Double
    Public time_over20 As Double
    Public time_20 As Double
    Public detected As Boolean = False

    Public ENa As Double          'Nernst Reversal potential 
    Public EK As Double
    Public ECa As Double
    Public ENaCa As Double      'mV, Reversal potential of INaCa (INCX)
    Public ENaK As Double        'mV, Reversal potential of INaK

    Public GHKK As Double             'mM, voltage dependency of K flux  defined by the Goldman, Hodgkin, Katz Equation
    Public GHKNa As Double           'mM, voltage dependency of Na flux  defined by the constantfield theory
    Public GHKCa As Double           'mM, voltage dependency of Ca flux defined by the constantfield theory

    '***************************Ion concentrations ******************************

    Public Ke As Double            'mM, External K   concentration,   [K]e
    Public Nae As Double          'mM, External Na concentration,   [Na]e
    Public Cae As Double           'mM, External Ca concentration,   [Ca]e
    Public newKe As Double       'new External K   concentration,   [K]e only used when [X]out is changed
    Public newNae As Double     'new External Na concentration,   [Na]e
    Public newCae As Double      'new External Ca concentration,   [Ca]e

    Public Ki As Double             'mM, intracellular K concentration,   [K]i
    Public Nai As Double           'mM, intracellular Na concentration, [Na]i
    Public dKidt As Double        'mM/msec
    Public dNaidt As Double      'mM/msec

    Public totCa_cyt As Double        'mM, [Catotal]cytosol
    Public dtotCa_cytdt As Double   'mM/ms, rate of change 
    Public Caf_cyt As Double           'Free Ca concentration in cytoplasm, [Ca2+]i

    Public Ca_SRup As Double           'mM, Ca concentration in the SR uptake site
    Public dCa_SRupdt As Double      'mM/ ms, rate of change
    Public Caf_SRrel As Double          'mM, free Ca concentration in the SR releasing site
    Public totCa_SRrel As Double       'total mM of [Ca]SRrl
    Public dtotCa_SRreldt As Double  'mM/ms, rate of change of total [Ca]SRrl


    Public Ggap As Double                          'External current in the RC membrane model. This is also applicable for the clamp current.
    Public IgapK As Double                           'gap junctional current carried by K
    Public IextK As Double                          'External current carried by K in the RC membrane model. This is also applicable for the clamp current.
    'Public IgapNa As Double                           'gap junctional current carried by Na
    'Public IextNa As Double                          'External current carried by Na in the RC membrane model. This is also applicable for the clamp current.
    'Public Ipip As Double
    'Public onset As Double
    'Public offset As Double
    '*************************voltage clamp & Current clamp conditions ****************************
    'Public VConset As Integer          'Onset (msec) of the voltage clamp pulse
    'Public VCoffset As Integer          'Offset (msec) of the voltage clamp

    'Public HoldingPotential As Integer  'Holding potential for the voltage clamp
    'Public TestPotential As Double          'Test potential
    'Public TestpulseOnset As Integer     'Onset time (msec) of the test pulse
    'Public TestpulseOffset As Integer     'Offset time (msec) of the test pulse
    'Public TestpulseStep As Integer        'voltage interval between successive test pulse

    '*************************(4-E)  Ca buffer   & Ca transient **************
    'Public KmEGTA As Double = 0.00000001995     'mM, Dissociation const of EGTA, SCHOEN
    'Public KmTRPN As Double = 0.5 * 0.001            'mM, Dissociation const of TROPONIN Luo & Rudy, Circ Res,74:1071-1096,1994

    Public KmCMDN As Double = 2.38 * 0.001      'mM, Dissociation const of calmodulin
    Public CMDNalfa As Double = 100                   ' /mM/msec
    Public CMDNbeta As Double = 238 / 1000       '/msec

    Public KmCSQN As Double = 0.8           'mM, Dissociation const of CALSEQUESTRIN Luo & Rudy, Circ Res,74:1071-1096,1994

    'Public EGTApip As Double                     'Concentration of EGTA in the pipette solution
    'Public EGTAdif As Double = 0.002          'A non-realistic diffusion parameter for EGTA, only to avoid calculation error.
    'Public EGTAChange As Double              'Changes in the EGTA concentration during a TimeInterval when the pipette EGTA is changed.
    Public BathDiffusion As Double = 0.002      'To simulate ion exchange experiment in the recording batth

    Public EGTAt As Double                   'mM, Total concentration of EGTA, m
    'Public TRPNt As Double = 0.07                 'mM, Total concentration of TROPONIN
    Public CMDNt As Double = 0.05                 'mM, Total concentration of CALMODULIN
    Public CSQNt As Double = 5                   'mM, Total concentration of CALSEQUESTRIN

    'Public EGTAf As Double                     'mM, free EGTA
    Public TRPNf As Double                     'mM, TROPONIN, 
    Public CMDNf As Double                    'mM, CALMODULIN
    Public CSQNf As Double                     'mM, CALSEQUESTRIN

    '*******************  (5-D)  net ION flux to estimate cellular concentration *****
    Public Itot As Double              '= ILB + ICaL + ICaT + Ito + IRK1 + IKr + IKs + INa + INaCa + INaK + Iext

    Public ICanet As Double       '= IbCa + ICaLCa + ICaT - 2 * INaCa - IrelRyR + IsrCaPup
    Public IKnet As Double          '= ICaLK + IKr + ITOK + IKsK + INACK + IRK1 + IKplateau + ILNSCK - 2 * INaK
    Public INanet As Double        '= ILNSCNA + ICaLNa + ITONa + INACNA + IKsNa + 3 * INaCa + 3 * INaK
    Public ILB As Double             'Sum of background currents

    '################ Parameters determining intracelluler ATP concentrations  ##################
    Public ATP As Double                                 'mM, intracelluler ATP concentration
    Public dATPdt As Double                            'mM/ms

    Public ATPbyINaK As Double                      'd[ATP] by Na-K pump
    Public ATPbyIsrCaPup As Double                'd[ATP] by srCa pump
    Public ATPbyCntr As Double                        'd[ATP] by Contraction
    Public SumATPbyPumps As Double             'Sum of ATP used by Na-K and Ca pumps
    Public SumATPbyCntr As Double                 'Sum of ATP used by Contraction

    Public ADP As Double                                     'intracelluler ADP concentration in mM
    Public alfaADP As Double = 0.0015                 'msec-1  rate of ADP --> ATP
    Public effiADP As Double                                'to simulate the intevention of the ATP production
    Public NucleotideTotal As Double = 5             'Total of ADP and ATP
    Public CellENGfromSumATPbyCntr As Double   'energy produced by ATP hydrolysis

    '################   (5-G) Parameters determining Energy consumptopn from Force-Length Area(FLA) #############
    Public minL As Double                   'Minimum length of half sarcomere length during one cycle of  contraction.
    Public maxL As Double                   'Maximum length of half sarcomere length during one cycle of  contractio
    Public length As Single
    Public ENGbyFLA As Double          'force-length area as an index of Energy consumption by contraction
    Public SumENGbyFLA As Double   'Energy consumption caluculated from FLA
    Public CellENGbyFLA As Double     'Energy consumption/Cell caluculated from FLA(pJ/cell)
    Public StepSL As Double                  'This is used when determining the new halfsarcomere length.

    '****************   (3-B)others for programming ***********
    Public N As Double                                'counter for general use.
    Public Denomi As Double                       'TimeInterval

    Public MaxINa As Double                        'maximum amplitude of the Na current
    Public MaxICaL As Double                        'maximum amplitude of the Na current
    Public MaxICaT As Double                        'maximum amplitude of the Na current
    Public MaxIha As Double                        'maximum amplitude of the Na current
    Public MaxI As Double                        'maximum amplitude of the Na current
    Public MaxIK1 As Double                        'maximum amplitude of the Na current
    Public MaxIKr As Double                        'maximum amplitude of the Na current
    Public MaxIKs As Double                        'maximum amplitude of the Na current
    Public MaxIst As Double                        'maximum amplitude of the Na current
    Public MaxIto As Double                        'maximum amplitude of the Na current
    Public MaxIKACh As Double                        'maximum amplitude of the Na current
    Public MaxINaK As Double                        'maximum amplitude of the Na current
    Public MaxINaCa As Double                        'maximum amplitude of the Na current
    Public MaxIKATP As Double                        'maximum amplitude of the Na current
    Public MaxILCCA As Double                        'maximum amplitude of the Na current
    Public MaxItot As Double                        'maximum amplitude of the Na current
    Public MaxIbNSC As Double                        'maximum amplitude of the Na current

    Public MaxSRca As Double                         'maximum [Ca]SRrl
    Public MaxRateOfRise As Double               'to calculate the maximum rate of rise (Na channel)
    Public DeltaItot As Double                         'maximum amplitude of the total current
    Public HR As Double                                   'min-1, Heart rate

    Public Datasize As Long            'Number of Data points to store, also determines stimulus frequency

    '*****************************      Technical parameters for calculation *****************************
    Public Nvar As Long = 71                         'number of data memory array   
    Public Mery(Nvar) As Double                   'memory array

    Public MemorySize As Long = 400000  '100100     '1 / timeinterval = 50,    timeinterval = 0.02
    Public EndAddress As Long                      'Last address of the memory array written.

    Public LeadPotential(MemorySize) As Double                        'memory array for print out

    Public temp As Double
    Public K As Integer
    Public maxtime(10) As Single

    '***********************  sub Class for individual functional components  **************************
    Public Ipip As Double                          'pipette current in the RC membrane model. This is also applicable for the clamp current.
    Public ampIpip As Double                    'pA, stimulus current amplitude 
    Public onset As Double = 50
    Public offset As Double = 53
    Public Sub calculateIext(ByVal time As Double)      'current applied from an external source 
        If time > onset And time < offset Then
            Ipip = ampIpip
        Else
            Ipip = 0
        End If
    End Sub

    Public INa_Na As Double                   'INa Na component
    Public INa_K As Double                     ' INa K component
    Public INa As Double                           '= INACNA + INACK
    ' ********************** Shirokov four state model **********************

    Public PNa As Double = 1041 * 1.5      ' / Cm                 'sodium current   limiting conductance
    Public RGKNA As Double = 0.1              'Ratio for K/Na
    Public kCO_Na As Double              'alfa m of Mitsuiye model is used for the four state Shirokov model
    Public kOC_Na As Double              'beta m of Mitsuiye model
    Public kOI_Na As Double                'beta i of Mitsuiye model
    Public kIO_Na As Double
    Public kIR_Na As Double
    Public kRI_Na As Double
    Public kCR_Na As Double
    Public kRC_Na As Double

    Public C_Na As Double        '   INa
    Public dC_Nadt As Double
    Public O_Na As Double
    Public dO_Nadt As Double
    Public I_Na As Double
    Public dI_Nadt As Double
    Public R_Na As Double

    Public kusIb_Na As Double               'ultra slow gate of two state model
    Public kusIf_Na As Double                'ultra slow gate of two state model
    Public us_Na As Double                 'ultra-slow gate (inactivation)
    Public dus_Nadt As Double

    Public Denominator As Double                 'avoid denominator = 0
    Public dt As Double = 0.02              'ms, Time interval for integration

    Public Sub calculateINa()
        '*************** INa  ***************
        kCO_Na = 1 / (0.1027 * Math.Exp((Vm) / -8) + 0.25 * Math.Exp((Vm) / -50))
        kOC_Na = 1 / (26 * Math.Exp((Vm) / 17) + 0.02 * Math.Exp((Vm) / 800))
        kRI_Na = 1 / (0.0001027 * Math.Exp((Vm) / -8) + 5 * Math.Exp((Vm) / -400))
        kIR_Na = 1 / (1300 * Math.Exp((Vm) / 20) + 0.04 * Math.Exp((Vm) / 800))
        kOI_Na = 1 / (0.8 * Math.Exp(Vm / -400))
        kIO_Na = 0.0000875
        kCR_Na = 0.01 / (1 + kIO_Na * kOC_Na * kRI_Na / kOI_Na / kCO_Na / kIR_Na)                'microscopic reversibility
        kRC_Na = 0.01 - kCR_Na

        kusIf_Na = 1 / (9000000000.0# * Math.Exp(Vm / 5) + 8000 * Math.Exp(Vm / 100))        'INa very slow inactivation alfa
        kusIb_Na = 1 / (0.014 * Math.Exp(Vm / -5) + 4000 * Math.Exp(Vm / -100))                   'INa very slow inactivation beta

        R_Na = 1 - C_Na - O_Na - I_Na
        dC_Nadt = kOC_Na * O_Na + kRC_Na * R_Na - (kCO_Na + kCR_Na) * C_Na
        dO_Nadt = kCO_Na * C_Na + kIO_Na * I_Na - (kOC_Na + kOI_Na) * O_Na
        dI_Nadt = kOI_Na * O_Na + kRI_Na * R_Na - (kIO_Na + kIR_Na) * I_Na

        dus_Nadt = kusIf_Na * (1 - us_Na) - kusIb_Na * us_Na

        Dim pOpen As Double = O_Na * us_Na
        INa_Na = PNa * GHKNa * pOpen
        INa_K = RGKNA * PNa * GHKK * pOpen
        INa = INa_Na + INa_K

    End Sub

    Public ICaL As Double                     'ICaL total
    Public PCaL As Double                     '20  Ca Permeability factor for L-Ca channel
    Public maxPCaL As Double = 1585.5      '/ Cm
    '20  Ca Permeability factor for L-Ca channel
    Public fPCaL As Double                   'dependence of L-Ca channel on [ATP]

    Public ICaLCa As Double                   'current component of ICaL carried by Ca
    Public ICaLK As Double                    'current component of ICaL carried by K
    Public ICaLNa As Double                   'current component of ICaL carried by Na
    Public ICaLCainf As Double                'fully activated Ca component
    Public ICaLKinf As Double                 'fully activated K component
    Public ICaLNainf As Double                ' fully activated Na component

    Public RCaLK As Double = 0.000733       '0.2% of PCaL
    Public RCaLNa As Double = 0.000037     'relative Na permeability ICaL

    Public SingleICaL As Double           'single channel current
    Public srkCabindsite As Double      'Ca concentration at the binding site of Ca gate

    Public kOvC_CaL As Double                 'voltage gate, rate constants
    Public kvCO_CaL As Double
    Public kIR_CaL As Double
    Public kRI_CaL As Double
    Public kOI_CaL As Double
    Public kIO_CaL As Double
    Public kvCR_CaL As Double
    Public kRvC_CaL As Double
    Public kslowb_CaL As Double
    Public kslowf_CaL As Double

    Public kCU_CaL As Double              'Ca-dependent gate, rate constants
    Public kUC_CaL As Double
    Public kCCaUCa_CaL As Double
    Public kUCaCCa_CaL As Double
    Public kCCCa_CaL As Double
    Public kCCaC_CaL As Double
    Public kUUCa_CaL As Double
    Public kUCaU_CaL As Double

    Public Vshift_CaL As Double                   'voltage shift of the activation parameter d. used to s

    Public vC_CaL As Double       'state C
    Public dvC_CaLdt As Double
    Public O_CaL As Double         'state O
    Public dO_CaLdt As Double
    Public I_CaL As Double           'state I
    Public dI_CaLdt As Double
    Public R_CaL As Double           'state R

    Public U_CaL As Double             'Ca-gate of ICaL
    Public dU_CaLdt As Double
    Public UCa_CaL As Double
    Public dUCa_CaLdt As Double
    Public cC_CaL As Double
    Public dcC_CaLdt As Double
    Public CCa_CaL As Double

    Public us_CaL As Double            'ultra-slow gate
    Public dus_CaLdt As Double

    Public Sub calculateICaL()
        ''************ ICA L  rates of the voltage dependent gate **************************
        kvCO_CaL = 1 / (0.27 * Math.Exp((Vm - Vshift_CaL) / -5.9) + 1.5 * Math.Exp((Vm - Vshift_CaL) / -65))
        kOvC_CaL = 1 / (480 * Math.Exp((Vm - Vshift_CaL) / 7) + 2.2 * Math.Exp((Vm - Vshift_CaL) / 65))
        kOI_CaL = 0.004
        kIO_CaL = 0.001
        kIR_CaL = 1 / (2200000 * Math.Exp((Vm - Vshift_CaL) / 7.4) + 11 * Math.Exp((Vm - Vshift_CaL) / 100))
        kRI_CaL = 1 / (0.0018 * Math.Exp((Vm - Vshift_CaL) / -7.4) + 2 * Math.Exp((Vm - Vshift_CaL) / -100))
        kvCR_CaL = 0.04 / (1 + kIO_CaL * kOvC_CaL * kRI_CaL / kOI_CaL / kvCO_CaL / kIR_CaL) 'microscopic reversibility
        kRvC_CaL = 0.04 - kvCR_CaL

        kslowf_CaL = 1 / (250000 * Math.Exp(Vm / 9) + 58 * Math.Exp(Vm / 65))     'ultra slow gate
        kslowb_CaL = 1 / (1800 * Math.Exp(Vm / -14) + 66 * Math.Exp(Vm / -65))    'ultra slow gate

        R_CaL = 1 - vC_CaL - O_CaL - I_CaL     'mass conservation
        dvC_CaLdt = kOvC_CaL * O_CaL + kRvC_CaL * R_CaL - (kvCO_CaL + kvCR_CaL) * vC_CaL
        dO_CaLdt = kIO_CaL * I_CaL + kvCO_CaL * vC_CaL - (kOI_CaL + kOvC_CaL) * O_CaL
        dI_CaLdt = kRI_CaL * R_CaL + kOI_CaL * O_CaL - (kIR_CaL + kIO_CaL) * I_CaL

        ''************ ICA L  rates of the Ca dependent gate **************************
        Denominator = (1 - Math.Exp(-Vm / RTF2))
        If Denominator = 0 Then
            SingleICaL = 0.005063 * Cae * 12.63
        Else
            SingleICaL = 0.005063 * Cae * Vm * Math.Exp(-Vm / RTF2) / Denominator
        End If
        'SingleICaL = 0.0676 * GHKCa
        srkCabindsite = Caf_cyt + 0.3 * SingleICaL         'Concentration at the Ca binding site for the Ca-mediated inactivation

        kCU_CaL = 0.143                                          '0.1
        kUC_CaL = 0.35                                            '1
        kCCaUCa_CaL = 0.0003                             '0.001
        kUCaCCa_CaL = 0.35                                 '1
        kCCCa_CaL = 6.954                                     ' original value = 1    Calcium binding rate  /mM/msec
        kCCaC_CaL = 0.0042                                 '0.005
        kUUCa_CaL = kCCCa_CaL                           'Calcium binding rate  /mM/msec
        kUCaU_CaL = kCCaC_CaL * kCU_CaL * kUCaCCa_CaL / kUC_CaL / kCCaUCa_CaL  'microscopic reversibility

        '***************  rate of gating variables **********************

        CCa_CaL = 1 - U_CaL - UCa_CaL - cC_CaL    'mass conservation
        Dim Calcium_U As Double = srkCabindsite * O_CaL + Caf_cyt * (1 - O_CaL)   'Ca at the nanodomain when Ca gate is open
        Dim Calcium_C As Double = srkCabindsite * O_CaL                                          'Ca at the nano domain when Ca gate is close
        dU_CaLdt = kCU_CaL * cC_CaL + kUCaU_CaL * UCa_CaL - (kUC_CaL + kUUCa_CaL * Calcium_U) * U_CaL
        dUCa_CaLdt = kUUCa_CaL * Calcium_U * U_CaL + kCCaUCa_CaL * CCa_CaL - (kUCaU_CaL + kUCaCCa_CaL) * UCa_CaL
        dcC_CaLdt = kUC_CaL * U_CaL + kCCaC_CaL * CCa_CaL - (kCU_CaL + kCCCa_CaL * Calcium_C) * cC_CaL

        dus_CaLdt = kslowf_CaL * (1 - us_CaL) - kslowb_CaL * us_CaL

        '****************** Current amplitudes **************************
        PCaL = maxPCaL / (1 + (1.4 / ATP) ^ 3)
        ICaLCainf = 2 * PCaL * GHKCa
        ICaLKinf = RCaLK * PCaL * GHKK
        ICaLNainf = RCaLNa * PCaL * GHKNa

        Dim pOpen As Double = O_CaL * (U_CaL + UCa_CaL) * us_CaL
        ICaLCa = ICaLCainf * pOpen
        ICaLK = ICaLKinf * pOpen
        ICaLNa = ICaLNainf * pOpen

        ICaL = ICaLCa + ICaLK + ICaLNa

    End Sub

    Public ICaTInf As Double
    Public ICaT As Single
    Public PCaT As Double = 111.4 / Cm                    'T-type Ca channel amplitude
    Public dalfa_CaT As Double
    Public dbeta_CaT As Double
    Public falfa_CaT As Double
    Public fbeta_CaT As Double

    Public d_CaT As Double            'ICaT
    Public dd_CaTdt As Double
    Public f_CaT As Double
    Public df_CaTdt As Double

    Public Sub calculateICaT()
        '************ ICAT  CURRENT *******************************
        dalfa_CaT = 1 / (0.019 * Math.Exp(Vm / -5.6) + 0.82 * Math.Exp(Vm / -250))
        dbeta_CaT = 1 / (40 * Math.Exp(Vm / 6.3) + 1.5 * Math.Exp(Vm / 10000))
        falfa_CaT = 1 / (62000 * Math.Exp(Vm / 10.1) + 30 * Math.Exp(Vm / 3000))
        fbeta_CaT = 1 / (0.0006 * Math.Exp(Vm / -6.7) + 1.2 * Math.Exp(Vm / -25))

        dd_CaTdt = dalfa_CaT * (1 - d_CaT) - dbeta_CaT * d_CaT
        df_CaTdt = falfa_CaT * (1 - f_CaT) - fbeta_CaT * f_CaT

        ICaTInf = 2 * PCaT * GHKCa
        ICaT = ICaTInf * d_CaT * f_CaT
    End Sub

    Public Ist As Double                      'Ist = IstNa + IstK
    Public IstNa As Double                    'Na component
    Public IstK As Double                     'K component
    Public IstNainf As Double
    Public PstNa As Double = 0                   'Na conductance
    Public IstKinf As Single
    Public RstK As Double = 0                   'K conductance

    Public vShift_st As Double              ' to shift voltage gate
    Public alfad_st As Double
    Public betad_st As Double
    Public alfaf_st As Double
    Public betaf_st As Double
    Public alfaus_st As Double
    Public betaus_st As Double

    Public d_st As Double                   'Ist
    Public f_st As Double
    Public us_st As Double
    Public dd_stdt As Double
    Public df_stdt As Double
    Public dus_stdt As Double

    Public Sub calculateIst()
        '***********************   Ist  ***********************************
        alfad_st = 1 / (0.15 * Math.Exp((Vm - vShift_st) / -11) + 0.2 * Math.Exp((Vm - vShift_st) / -700))            'from rat data, alpha, activation parameter for IST
        betad_st = 1 / (16 * Math.Exp((Vm - vShift_st) / 8) + 15 * Math.Exp((Vm - vShift_st) / 50))                        'from rat data, beta

        alfaf_st = 1 / (3100 * Math.Exp((Vm - vShift_st) / 13) + 700 * Math.Exp((Vm - vShift_st) / 70))             'from rat data, alpha, activation parameter for IST
        betaf_st = 1 / (95 * Math.Exp((Vm - vShift_st) / -10) + 50 * Math.Exp((Vm - vShift_st) / -700)) + Caf_cyt * 2.5 / (1 + Math.Exp(Vm / -5))                       'from rat data, beta

        alfaus_st = 1 / (400000 * Math.Exp(Vm / 9) + 60 * Math.Exp(Vm / 65))
        betaus_st = 1 / (700 * Math.Exp(Vm / -14) + 60 * Math.Exp(Vm / -65))

        dd_stdt = alfad_st * (1 - d_st) - betad_st * d_st
        df_stdt = alfaf_st * (1 - f_st) - betaf_st * f_st
        dus_stdt = alfaus_st * (1 - us_st) - betaus_st * us_st


        IstNainf = PstNa * GHKNa
        IstKinf = RstK * PstNa * GHKK
        Dim pOpen As Double = d_st * f_st * us_st
        IstNa = IstNainf * pOpen
        IstK = IstKinf * pOpen
        Ist = IstNa + IstK
    End Sub

    Public Iha As Double                      'total = IhaK + IhaNa
    Public IhaK As Double                     'K component of Iha
    Public IhaNa As Double                    'Na component of Iha
    Public PKha As Double = 0                  'Amplitude factor for the K comp. Iha
    Public PNaha As Double = 0                   'Amplitude factor for the Na comp. Iha

    Public alfaf_ha As Double
    Public betaf_ha As Double                    'closing rate
    Public myuf_ha As Double                     'activation rate from C1 to C2
    Public Lamdaf_ha As Single

    Public vshift_ha As Double                   'voltage shift of the Iha kinetics.

    Public C1_ha As Double              'Iha
    Public C2_ha As Double
    Public O1_ha As Double
    Public O2_ha As Double
    Public dC1_hadt As Double
    Public dC2_hadt As Double
    Public dO1_hadt As Double
    Public dO2_hadt As Double

    Public Sub calculateIha()
        '************* Iha  CURRENT  ******************
        alfaf_ha = 1 / (3500 * Math.Exp((Vm - vshift_ha) / 16.8) + 0.3 * Math.Exp((Vm - vshift_ha) / 400))                                        'slightly faster than Maruoka 1994
        betaf_ha = 1 / (4 * Math.Exp((Vm - vshift_ha) / -14) + 2 * Math.Exp((Vm - vshift_ha) / -400))
        myuf_ha = 1 / (45000000 * Math.Exp((Vm - vshift_ha) / 8) + 500 * Math.Exp((Vm - vshift_ha) / 200))
        Lamdaf_ha = 1 / (10.5 * Math.Exp((Vm - vshift_ha) / -16.4) + 0.4 * Math.Exp((Vm - vshift_ha) / -400))

        dC1_hadt = Lamdaf_ha * C2_ha - myuf_ha * C1_ha                                                         'C1 state of If
        dC2_hadt = myuf_ha * C1_ha + betaf_ha * O1_ha - (Lamdaf_ha + alfaf_ha) * C2_ha       'C2 state of if channel
        dO1_hadt = alfaf_ha * C2_ha + betaf_ha * O2_ha - (betaf_ha + alfaf_ha) * O1_ha            'O1 state
        dO2_hadt = alfaf_ha * O1_ha + betaf_ha * (1 - C1_ha - C2_ha - O1_ha - O2_ha) - (alfaf_ha + betaf_ha) * O2_ha  'O2 state

        IhaK = PKha * GHKK * (1 - C1_ha - C2_ha)
        IhaNa = PNaha * GHKNa * (1 - C1_ha - C2_ha)
        Iha = IhaNa + IhaK

    End Sub

    Public IK1 As Single
    Public maxGIK1 As Double = 18.382      '/ Cm
    Public GIK1 As Double                     '0.1

    Public myu_Mg As Double                 'M^-1,msec^-1  single channel data maximum slope
    Public lamda_Mg As Double              '120, 'adjust half block = -50 mV
    Public fracB As Double                       'fraction of Mg-blocked
    Public fracO As Double                       'fraction of Mg-unblocked component
    Public bkspm_K1 As Double
    Public aKspm_K1 As Double

    Public O_K1 As Double                 'IK1
    Public dO_K1dt As Double

    Public Sub calculateIK1()
        '******** Ko-dependence of IKr channel  *****************
        aKspm_K1 = 1 / (8000 * Math.Exp((Vm - EK - 97) / 8.5) + 7 * Math.Exp((Vm - EK - 97) / 300))
        bkspm_K1 = 1 / (0.00014 * Math.Exp((Vm - EK - 97) / -9.1) + 0.2 * Math.Exp((Vm - EK - 97) / -500))

        myu_Mg = 0.75 * Math.Exp(0.035 * (Vm - EK - 10)) / (1 + Math.Exp(0.015 * (Vm - EK - 140)))
        lamda_Mg = 3 * Math.Exp(-0.048 * (Vm - EK - 10)) * (1 + Math.Exp(0.064 * (Vm - EK - 38))) / (1 + Math.Exp(0.03 * (Vm - EK - 70)))

        fracB = myu_Mg / (myu_Mg + lamda_Mg)         'Mg-block確率
        fracO = lamda_Mg / (myu_Mg + lamda_Mg)       'open 確率

        dO_K1dt = aKspm_K1 * (1 - O_K1) - bkspm_K1 * O_K1 * fracO ^ 4     'fraction of channels , which are not blocked by spermine

        GIK1 = maxGIK1 * (Ke / 5.4) ^ 0.4

        IK1 = GIK1 * (Vm - EK) * (fracO ^ 4 + 2 / 3 * 4 * fracO ^ 3 * fracB + 1 / 3 * 6 * fracO ^ 2 * fracB ^ 2) * O_K1

    End Sub


    Public IKr As Double                      'IK,r
    Public GKr As Double
    Public maxGKr As Double = 0.10395 * 0.5 * Cm        '*0.4    0.099  original from Ono for the SA node cell. Conductance factor

    Public y1_Kr As Double
    Public y2_Kr As Double
    Public y3_Kr As Double
    Public dy1_Krdt As Double
    Public dy2_Krdt As Double
    Public dy3_Krdt As Double

    Public alfay1_Kr As Double
    Public betay1_Kr As Double
    Public alfay2_Kr As Double
    Public betay2_Kr As Double
    Public alfay3_Kr As Double
    Public betay3_Kr As Double

    Public Sub calculateIKr()
        '******** Ko-dependence of IKr channel  *****************


        alfay1_Kr = 1 / (20 * Math.Exp(Vm / -11.5) + 5 * Math.Exp(Vm / -300))                   'These equations can simulate the original data.
        betay1_Kr = 1 / (160 * Math.Exp(Vm / 28) + 200 * Math.Exp(Vm / 1000)) + 1 / (2500 * Math.Exp(Vm / 20))

        alfay2_Kr = 1 / (200 * Math.Exp(Vm / -13) + 20 * Math.Exp(Vm / -300))
        betay2_Kr = 1 / (1600 * Math.Exp(Vm / 28) + 2000 * Math.Exp(Vm / 1000)) + 1 / (10000 * Math.Exp(Vm / 20))

        alfay3_Kr = 1 / (10 * Math.Exp(Vm / 17) + 2.5 * Math.Exp(Vm / 300))
        betay3_Kr = 1 / (0.35 * Math.Exp(Vm / -17) + 2 * Math.Exp(Vm / -150))
        dy1_Krdt = alfay1_Kr * (1 - y1_Kr) - betay1_Kr * y1_Kr
        dy2_Krdt = alfay2_Kr * (1 - y2_Kr) - betay2_Kr * y2_Kr
        dy3_Krdt = alfay3_Kr * (1 - y3_Kr) - betay3_Kr * y3_Kr

        GKr = maxGKr * (Ke / 5.4) ^ 0.2               '0.35         '0.099  original from Ono. Conductance factor for IK,rapid? conductance increases with higher Ke.
        IKr = GKr * (Vm - EK) * (0.6 * y1_Kr + 0.4 * y2_Kr) * y3_Kr

    End Sub

    Public IKs As Double                      '= IKsK + IKsNa
    Public IKsK As Single
    Public IKsNa As Double                    'IKs Na component
    Public PIKsK As Double = 0.027       '/ Cm                   'IKs Amplitude factor , ZengJL et al, 1995, CR 77:140-152
    Public RGIKsNa As Double = 0.04              'To give a reversal potential at () mV

    Public alfav_Ks As Double
    Public betav_Ks As Double
    Public alfac_Ks As Double
    Public betac_Ks As Double

    Public y1_Ks As Double               'IKs
    Public y2_Ks As Double
    Public dy1_Ksdt As Double
    Public dy2_Ksdt As Double

    Public Sub calculateIKs()
        ''************* IKs CURRENT  *****************
        alfav_Ks = 1 / (85 * Math.Exp(Vm / -10.5) + 370 * Math.Exp(Vm / -62))   'voltage-gate of IKs
        betav_Ks = 1 / (1450 * Math.Exp(Vm / 20) + 260 * Math.Exp(Vm / 100))

        alfac_Ks = Caf_cyt * 3.7                  'Ca-gate of IKs
        betac_Ks = 0.004444

        dy1_Ksdt = alfav_Ks * (1 - y1_Ks) - betav_Ks * y1_Ks
        dy2_Ksdt = alfac_Ks * (1 - y2_Ks) - betac_Ks * y2_Ks


        IKsK = PIKsK * GHKK * y1_Ks ^ 2 * (0.9 * y2_Ks + 0.1)
        IKsNa = RGIKsNa * PIKsK * GHKNa * y1_Ks ^ 2 * (0.9 * y2_Ks + 0.1)
        IKs = IKsK + IKsNa
    End Sub

    Public IKto As Double                      ' = ITOK + ITONa
    Public ItoK As Double                     'K component
    Public ItoNa As Double                    'Na component
    Public PK_Kto As Double = 0.012       '/ Cm
    Public RPNa_Kto As Double = PK_Kto * 0.09                  'Na component

    Public y1_Kto As Double
    Public y2_Kto As Double
    Public dy1_Ktodt As Double
    Public dy2_Ktodt As Double

    Public alfa1_Kto As Double
    Public beta1_Kto As Double
    Public alfa2_Kto As Double
    Public beta2_Kto As Double

    Public Sub calculateIKto()
        '************* Ito  CURRENT  *****************
        alfa1_Kto = 1 / (11 * Math.Exp(Vm / -28) + 0.2 * Math.Exp(Vm / -400))
        beta1_Kto = 1 / (4.4 * Math.Exp(Vm / 16) + 0.2 * Math.Exp(Vm / 500))
        alfa2_Kto = 0.0038 * Math.Exp((Vm + 13.5) / -11.3) / (1 + 0.051335 * Math.Exp((Vm + 13.5) / -11.3))
        beta2_Kto = 0.0038 * Math.Exp((Vm + 13.5) / 11.3) / (1 + 0.067083 * Math.Exp((Vm + 13.5) / 11.3))

        dy1_Ktodt = alfa1_Kto * (1 - y1_Kto) - beta1_Kto * y1_Kto
        dy2_Ktodt = alfa2_Kto * (1 - y2_Kto) - beta2_Kto * y2_Kto

        Dim pOpen As Double = y1_Kto ^ 3 * y2_Kto
        ItoK = PK_Kto * GHKK * pOpen
        ItoNa = RPNa_Kto * GHKNa * pOpen
        IKto = ItoK + ItoNa

    End Sub

    Public IKACh As Double
    Public ACh As Single
    Public newACh As Single
    Public GKACh As Double = 0.135       '/ Cm

    Public O_ACh As Double
    Public dO_AChdt As Double
    Public alfa_ACh As Double
    Public beta_ACh As Double

    Public Sub calculateIKACh()
        '************* ACH CURRENT  *****************
        If ACh = 0 Then
            alfa_ACh = 0.0002475                             'spontaneous opening   /msec"
        Else
            alfa_ACh = 0.01232 / (1 + 0.0042 / ACh) + 0.0002475
        End If
        beta_ACh = 0.01 * Math.Exp(0.0133 * (Vm + 40))           'beta of IKACh

        dO_AChdt = alfa_ACh * (1 - O_ACh) - beta_ACh * (O_ACh)

        IKACh = GKACh * (Vm - EK) / (1 + Math.Exp((Vm + 20) / 20)) * O_ACh

    End Sub

    Public IKATP As Double                 '##################   IKATP    ###################
    Public N_KATP As Double = 309       ' / Cm              'total number of KATP channels      ref. Kakei et al 1985
    Public rKATP As Double                    'single channel conductance (pS)
    Public PoK_ATP As Double = 0.8        'maximum open probability of KATP channel

    Public Sub calculateIKATP()
        '*********************** KATP currents ******************************
        rKATP = 23.6 * Ke ^ 0.24 / 1000                                'ATP sensitive K channel
        IKATP = N_KATP * PoK_ATP * rKATP * (Vm - EK) / (1 + (ATP / 0.1) ^ 2)
    End Sub

    Public INaK As Double                     'INaK
    Public INaKmax As Double = 22 * 0.8 * Cm      'SAKAI ET AL. J.P. 490:51-62
    Public KmNaintNaK As Double = 4.05      'mM, Collins, Somlyo, & Hilgemann, 1992, J. Physiol. 454:27-57 Giant patch method
    Public KmKkintNaK As Double = 32.88       'mM adjusted on model
    Public KmNaextNaK As Double = 69.8       'mM adjusted on the model
    Public KmKkextNaK As Double = 0.258     'mM Nakao & Gadsby 1989 J. Physiol. 94:539-565
    Public nHNaNaK As Double = 1.06        'experimental value, Collins, Somlyo, & Hilgemann, 1992, J. Physiol. 454:27-57 Giant patch method
    Public nHKkNaK As Double = 1.12        'Nakao & Gadsby 1989 J. Physiol. 94:539-565

    Public ramdaNaK As Double = 0.82     'Nakao & Gadsby 1989 J. Physiol. 94:539-565   depth of the access channel
    Public Naeff As Double                         'effective Na concentration at the bottom of the access channel

    Public k1_NaK As Double = 370 / 1000      '490 Nakao & Gadsby 1989 J. Physiol. 94:539-565 msec-1
    Public k2_NaK As Double = 40 / 1000       'Nakao & Gadsby 1989 J. Physiol. 94:539-565  msec-1
    Public k3_NaK As Double = 10 / 1000       'Nakao & Gadsby 1989 J. Physiol. 94:539-565     msec-1
    Public k4_NaK As Double = 165 / 1000      'Nakao & Gadsby 1989 J. Physiol. 94:539-565  msec-1
    Public KmATPNaK As Double = 0.094      'Collins, Somlyo, & Hilgemann, 1992, J. Physiol. 454:27-57 Giant patch method

    Public proPiNa As Double              'probability of Na-Xi conformation
    Public proPiKk As Double              'probability of K-Xi conformation
    Public proPeNa As Double              'probability of Na-Xo conformation
    Public proPeKk As Double              'probability of K-Xo conformation

    Public alfa_NaK As Double
    Public beta_NaK As Double

    Public PXi_NaK As Double            'Na/K pump
    Public dPXi_NaKdt As Double

    Public Fouabain As Double = 1                  'inhibition by ouabain
    Public Couabain As Double                    'concentration of ouabain (mM)

    Public Sub calculateNaK()
        '***************************** Ion transporters **************************************************************
        proPiNa = 1 / (1 + (KmNaintNaK / Nai) ^ nHNaNaK * (1 + (Ki / KmKkintNaK) ^ nHKkNaK))
        proPiKk = 1 / (1 + (KmKkintNaK / Ki) ^ nHKkNaK * (1 + (Nai / KmNaintNaK) ^ nHNaNaK))

        Naeff = Nae * Math.Exp(-ramdaNaK * Vm / RTF)            'effective Na concentration at the bottom of the access channel
        proPeNa = 1 / (1 + (KmNaextNaK / Naeff) ^ nHNaNaK * (1 + (Ke / KmKkextNaK) ^ nHKkNaK))
        proPeKk = 1 / (1 + (KmKkextNaK / Ke) ^ nHKkNaK * (1 + (Naeff / KmNaextNaK) ^ nHNaNaK))

        alfa_NaK = (k2_NaK * proPeNa) + (k4_NaK * proPeKk)   'rate constant of the reduced two-states model
        beta_NaK = (k1_NaK * proPiNa * (1 / (1 + KmATPNaK / ATP))) + (k3_NaK * proPiKk)           'rate constant of the reduced two-states model

        dPXi_NaKdt = alfa_NaK * (1 - PXi_NaK) - beta_NaK * PXi_NaK

        INaK = INaKmax * Fouabain * (k1_NaK * proPiNa * PXi_NaK - k2_NaK * proPeNa * (1 - PXi_NaK))
    End Sub

    Public partition = 0.32

    Public INaCa As Double                    'Pingpon model
    Public INaCamax As Double = 6.81 * Cm           'pA/pF
    Public KNaI = 8.75                           'Dissociation const for Na at the inner side of the membrane
    Public KNaO = 87.5                        'Dissociation const for Na at the outer side of the membrane
    Public KCaI = 1.38 * 0.001            'Dissociation const for Ca at the inner side of the membrane
    Public KCaO = 1.38                        'Dissociation const for Ca at the outer side of the membrane
    Public PNaO As Double                    'Probability of Na binding facing outer surface of the membrane
    Public PNaI As Double                     'Probability of Na binding facing the inner surface
    Public PCaO As Double                   'Probability of Ca binding facing outer surface
    Public PCaI As Double                     'Probability of Ca binding facing inner surface

    Public k2_NaCa As Double                      'rate constant for state transition
    Public k1_NaCa As Double                      'rate constant for state transition
    Public k4_NaCa As Double                      'rate constant for state transition
    Public k3_NaCa As Double                      'rate constant for state transition

    Public alfa_NaCa As Double
    Public beta_NaCa As Double

    Public PXi_NaCa As Double          'NaCa exchange
    Public dPXi_NaCadt As Double

    Public Sub calculateNaCa()
        '************  Na/Ca EXCHANGE PINGPON ****************
        PNaO = 1 / (1 + ((KNaO / Nae) ^ 3) * (1 + Cae / KCaO))             'PROBABILITY OF Na BINDING
        PNaI = 1 / (1 + ((KNaI / Nai) ^ 3) * (1 + Caf_cyt / KCaI))            'PROBABILITY OF Na BINDING
        PCaO = 1 / (1 + (KCaO / Cae) * (1 + (Nae / KNaO) ^ 3))              'PROBABILITY OF Ca BINDING
        PCaI = 1 / (1 + (KCaI / Caf_cyt) * (1 + (Nai / KNaI) ^ 3))             'PROBABILITY OF Ca BINDING
        k1_NaCa = Math.Exp(partition * Vm / RTF) * PNaI
        k2_NaCa = Math.Exp((partition - 1) * Vm / RTF) * PNaO
        k3_NaCa = PCaI
        k4_NaCa = PCaO

        alfa_NaCa = k2_NaCa + k4_NaCa                                                   'RATE TO   PXI
        beta_NaCa = k1_NaCa + k3_NaCa                                                  'RATE FROM PXI

        dPXi_NaCadt = alfa_NaCa * (1 - PXi_NaCa) - beta_NaCa * PXi_NaCa

        INaCa = INaCamax * (k1_NaCa * PXi_NaCa - k2_NaCa * (1 - PXi_NaCa))

    End Sub

    Public IsrCaPup As Double                             'SRCA pump  current
    Public ISERCAMax As Double = 162500 * SRaFactor                'limiting value of the IsrCaP
    Public KmsrCaPup As Double = 0.08         'mM, 0.5  dissociation const for Ca binding within SR
    Public KmsrCaPcp As Double = 0.0008     'mM,  0.0003 dissociation const for Ca binding within cytoplasm
    Public k_srCaP As Double = 0.01                  '0.4   translocation constant.

    Public PXo_CaP As Double                         'probability of Ca binding at the cytoplasmic side
    Public Xi_CaP As Double                          'probability of Ca binding at the SR side
    Public k1_CaP As Double                            'rate constant for state transition
    Public k2_CaP As Double                            'rate constant for state transition
    Public k3_CaP As Double                             'rate constant for state transition
    Public k4_CaP As Double                             'rate constant for state transition
    Public KmATP_CaP As Double = 0.1      'model adjusted.

    Public alfa_CaP As Double
    Public beta_CaP As Double

    Public dPXi_CaPdt As Double
    Public PXi_CaP As Double

    Public Sub calculateSERCA()
        '************  SR Ca pump (Hilgemann-type) ****************
        PXo_CaP = 1 / (1 + KmsrCaPcp / Caf_cyt)           'PROBABILITY OF Ca BINDING on the cytoplasmic side
        Xi_CaP = 1 / (1 + KmsrCaPup / Ca_SRup)           'PROBABILITY OF Ca BINDING on the SR internal side

        k1_CaP = k_srCaP * Xi_CaP
        k2_CaP = PXo_CaP
        k3_CaP = (1 - Xi_CaP)
        k4_CaP = k_srCaP * (1 - PXo_CaP)

        alfa_CaP = k2_CaP + k4_CaP                                                       'RATE TO   PXI
        beta_CaP = k1_CaP + k3_CaP * (1 / (1 + KmATP_CaP / ATP))    'RATE FROM PXI

        dPXi_CaPdt = alfa_CaP * (1 - PXi_CaP) - beta_CaP * PXi_CaP

        IsrCaPup = ISERCAMax * (k2_CaP * (1 - PXi_CaP) - k1_CaP * PXi_CaP)  'Ca influx to SR gives an outward current.

    End Sub

    Public IrelRyR As Double                   'Current due to the Ca release from SR releasing site.
    Public PRyR As Double = 62000 * SRaFactor                    'pA/(mM/L) Amplitude factor for the RyR Ca release
    Public ActivationFactorRyR As Double = 150      'amplitude factor for the activation of RyR in relation to ICaL

    Public kOI_RyR As Single
    Public kCI_RyR As Double                     'transitions between the available and nonavailable states.
    Public kIC_RyR As Double                     'msec-1.mM-1  define transitions between the available and nonavailable states.
    Public kCO_RyR As Double                     'define rate of open transition

    Public C_RyR As Double                  'RyR
    Public dC_RyRdt As Double
    Public O_RyR As Double
    Public dO_RyRdt As Double
    Public I_RyR As Double

    Public Sub calculateRyR()
        '****** 3 states model of available , open and unavailable state *************
        kCO_RyR = SingleICaL * O_CaL * (U_CaL + UCa_CaL) * us_CaL * ActivationFactorRyR + 280000 * Caf_cyt ^ 2
        kOI_RyR = 0.1 / (1 + 0.36 / Caf_SRrel)                              '0.08   rate of inactivation.
        kIC_RyR = Caf_SRrel ^ 2 * 0.000377                                  'recovery from inactivated state to resting state
        kCI_RyR = 0.000849                                                       'inactivation from the resting state

        I_RyR = 1 - C_RyR - O_RyR
        dC_RyRdt = kIC_RyR * I_RyR - (kCI_RyR + kCO_RyR) * C_RyR         'RyR kinetics
        dO_RyRdt = kCO_RyR * C_RyR - kOI_RyR * O_RyR           '

        IrelRyR = PRyR * (Caf_SRrel - Caf_cyt) * O_RyR     'new RyR kinetics, 

    End Sub

    Public srTransfer As Double = 386 * SRaFactor          'Amplitude factor for Ca transfer from the uptake site to the releasing site
    Public IsrTransfer As Double           'current for Ca transfer from the SR uptake site to the releasing site.

    Public Sub calculateSRCatransfer()
        IsrTransfer = srTransfer * (Ca_SRup - Caf_SRrel)                           'diffusion between uptake site and releasing site
    End Sub

    Public IsrLeak As Double                  'Leak from the store SR
    Public GsrLeak As Double = 459 * SRaFactor             'conductance of Leak from the store SR

    Public Sub calculateSRCaleak()
        IsrLeak = GsrLeak * (Ca_SRup - Caf_cyt)              'Leak from the store SR
    End Sub

    Public maxP_Kpl As Double = 0.01386 * 0.4
    Public P_Kpl As Double                '
    Public IKpl As Double

    Public Sub calculateIKpl()
        'maxP_Kpl = 0.00011 / Cm
        ''******** Ko-dependence of IK plateau channel   *****************
        P_Kpl = maxP_Kpl * (5.4 / 5.4) ^ 0.16    'K Component,   conductance increases with higher Ke.

        If Vm = -3 Then                                                '******* outward rectifier *****
            IKpl = P_Kpl * GHKK * 13.0077
        Else
            IKpl = P_Kpl * GHKK * (Vm + 3) / (1 - Math.Exp(-(Vm + 3) / 13))
        End If
    End Sub

    Public Ibnsc As Double                      '=ILNSCNa + ILNSCK
    Public IbnscNa As Double               '= PILNSCNa * NaCF   Na component of the Non-selective Cation current
    Public IbnscK As Double                  '= PILNSCK * KCF
    Public P_bnscNa As Double = 0.14    '/ Cm        'pA/mM  background Na current
    Public P_bnscK As Double = P_bnscNa * 0.4              'pA/mM  background K current     'leak K component

    Public Sub calculateIbNSC()

        IbnscNa = P_bnscNa * GHKNa
        IbnscK = P_bnscK * GHKK
        Ibnsc = IbnscNa + IbnscK

    End Sub

    Public IbCa As Double                      '= PbCa * CaCF  '(Vmembrane - ECa)
    Public P_bCa As Double = 0.00728       '/ Cm                    '2.75   backgrund Ca conductance

    Public Sub calculateIbCa()
        IbCa = 2 * P_bCa * GHKCa
    End Sub

    Public ILCCaNa As Double                 'Na component of Leak cation current induced by the intracellular Ca Ehara, Noma & Ono, 1988.
    Public ILCCaK As Double                   'K component of Leak cation current induced by the intracellular Ca Ehara, Noma & Ono, 1988.
    Public ILCCa As Double                     'Leak cation current induced by the intracellular Ca Ehara, Noma & Ono, 1988.
    Public PLCCa As Double = 0.04        ' / Cm                   'pA/mM   conductance for the ILCCa
    Public Sub calculateIL_Ca()
        ILCCaNa = PLCCa * GHKNa / (1 + (0.0012 / Caf_cyt) ^ 3)    'Leak cation current induced by the intracellular Ca Ehara, Noma & Ono, 1988.
        ILCCaK = PLCCa * GHKK / (1 + (0.0012 / Caf_cyt) ^ 3)      'Leak cation current induced by the intracellular Ca Ehara, Noma & Ono, 1988.
        ILCCa = ILCCaNa + ILCCaK
    End Sub


    Public Sub CaBuffer()
        '*****************  for the cytoplasmic Ca *****************************************
        'N = 0
        'If totCa_cyt > 0.000001 Then    ' < 1 nM
        '    Do
        '        Caf_cyt = totCa_cyt / (1 + (CMDNf / KmCMDN) + (EGTAf / KmEGTA))

        '        EGTAf = EGTAt / (1 + (Caf_cyt / KmEGTA))
        '        CMDNf = CMDNt / (1 + (Caf_cyt / KmCMDN))
        '        N = N + 1
        '    Loop While N < 6           '10 cycles seems to be good enough.
        'End If
        Dim aa As Double = 1
        Dim bb As Double = CMDNt - totCa_cyt + KmCMDN                 '(Btotal - Catotal + KD)
        Dim cc As Double = -KmCMDN * totCa_cyt                                     'KD * Catotal
        Caf_cyt = (-bb + Math.Sqrt(bb * bb - 4 * aa * cc)) / (2 * aa)                 'solution of second-order equation
        '****************** for the Ca in the releaseing site SR **************************
        'N = 0
        'Do
        '    Caf_SRrel = totCa_SRrel / (1 + (CSQNf / KmCSQN))
        '    CSQNf = CSQNt / (1 + (Caf_SRrel / KmCSQN))
        '    N = N + 1
        'Loop While N < 6           '10 cycles seems to be good enough.
        aa = 1
        bb = CSQNt - totCa_SRrel + KmCSQN                 '(Btotal - Catotal + KD)
        cc = -KmCSQN * totCa_SRrel                                     'KD * Catotal
        Caf_SRrel = (-bb + Math.Sqrt(bb * bb - 4 * aa * cc)) / (2 * aa)                 'solution of second-order equation
    End Sub

    Public Sub Igapjunc(ByRef Igj As Double, ByVal gGap As Double, ByVal Vma As Double, ByVal Vmb As Double, ByVal Ka As Double, ByVal Kb As Double)
        Dim Ec As Double = RTF * Math.Log(Ka / Kb)  'calculate the equilibrium potential for Ka/Kb using the Nernst Eq.
        Igj = gGap * ((Vma - Vmb) + Ec)
    End Sub

    Public Iext As Double          'sum of the gap junctional current
    Public Sub Iexternal(ByRef Iext As Double, ByVal igja As Double, ByVal igjb As Double)
        Iext = -igja + igjb      'sum the two gap junctional currents on both sides for the current integration

    End Sub

    Public Sub Preparedydt(ByVal sweeptime As Double)

        CaBuffer()
        NernstGHK()                                          'Calculate Nernst equilibrium potential and GHK ion flux
        MembraneCurrent(sweeptime)             ' RateConstants()'determine TimeVariable(y,1) & TimeVariable(y,2)      CurrentAmplitude()'prepare for calculation of deltaE = dV/dt = -I / Cm
        MembranePotentials()
        SRCadynamics()
        Contraction()
        CellIonConc()

    End Sub

    Public Sub CellIonConc()
        IKnet = INa_K + ICaLK + IK1 + IKr + IKsK + ItoK + IKpl + IKATP + IbnscK + ILCCaK - 2 * INaK
        INanet = INa_Na + ICaLNa + ItoNa + IKsNa + IbnscNa + ILCCaNa + 3 * INaK + 3 * INaCa
        ICanet = ICaLCa + ICaT + IbCa - 2 * INaCa

        dKidt = -(IKnet + IextK + Ipip) / Vi / F     'cytoplasmic K concentration, Assumption: External current is carried by K.
        dNaidt = -INanet / Vi / F                       'cytoplasmic Na

        dtotCa_cytdt = -(ICanet - IrelRyR + IsrCaPup - IsrLeak) / Vi / F / 2 + CaTropC    'calculate rate of ion concentration change, total plasma Ca
        dtotCa_SRreldt = (IsrTransfer - IrelRyR) / 2 / Vrel / F                         'SR releasing site
        dCa_SRupdt = (IsrCaPup - IsrTransfer - IsrLeak) / 2 / Vup / F       'SR uptake site


    End Sub

    Public Sub MembranePotentials()
        ILB = IKpl + Ibnsc + IbCa + IKATP + ILCCa                                           'background current
        Itot = INa + ICaL + ICaT + IK1 + IKr + IKs + IKto + ILB + INaK + INaCa   'Total current
        dVmdt = -(Itot + IextK + Ipip) / Cm                                        'dV/dt for membrane potential
    End Sub

    Public Sub NernstGHK()
        ENa = RTF * Math.Log(Nae / Nai)          'Reversal potential for  Na
        EK = RTF * Math.Log(Ke / Ki)                 'Reversal potential for K
        ECa = RTF2 * Math.Log(Cae / Caf_cyt)      'Reversal potential for Ca
        ENaCa = 3 * ENa - 2 * ECa            'Reversal potential for the Na/Ca exchange
        ENaK = -200                                   'Measured from the model

        '******* constant field for Ca *********
        Denominator = (1 - Math.Exp(-Vm / RTF2))
        If Denominator = 0 Then
            GHKCa = 1
        Else
            GHKCa = (Vm / RTF2) / Denominator
        End If
        GHKCa = GHKCa * (Caf_cyt - Cae * Math.Exp(-Vm / RTF2))

        '******* constant field for K *********
        Denominator = (1 - Math.Exp(-Vm / RTF))
        If Denominator = 0 Then
            GHKK = 1
        Else
            GHKK = (Vm / RTF) / Denominator
        End If
        GHKK = GHKK * (Ki - Ke * Math.Exp(-Vm / RTF))

        '******* constant field for Na *********
        If Denominator = 0 Then
            GHKNa = 1
        Else
            GHKNa = (Vm / RTF) / Denominator
        End If
        GHKNa = GHKNa * (Nai - Nae * Math.Exp(-Vm / RTF))

    End Sub

    Public Sub MembraneCurrent(ByVal sweeptime As Double)
        'calculateIext(sweeptime)
        calculateINa()   'calculate INa
        calculateICaL()           'calculate ICaL
        calculateICaT()            'calculate ICaT
        'calculateIst()           'calculate Ist
        'calculateIha()           'calculate Iha
        calculateIK1()           'calculate IK1
        calculateIKs()           'calculate IKs
        calculateIKto()           'calculate Ito
        'calculateIKACh()           'calculate IKACH
        calculateIKr()               'calculate Ikr

        calculateIKpl()               ' IK plateau
        calculateIbNSC()
        calculateIbCa()
        calculateIL_Ca()

        calculateIKATP()           'calculate IKATP
        calculateNaK()           'calculate INak 
        calculateNaCa()           'calculate  INaCa


        '*******************  net ION flux through the surface membrane, to calculate ion concentrations****************************

    End Sub

    Public Sub SRCadynamics()
        '*********************** SR Ca kinetics ******************************
        calculateSERCA()           'calculate SR
        calculateRyR()           'calculate RyR
        calculateSRCatransfer()
        calculateSRCaleak()

    End Sub


    Public Sub EulerMethod(ByVal dt As Double)

        Vm = Vm + dVmdt * dt

        us_Na = us_Na + dus_Nadt * dt
        C_Na = C_Na + dC_Nadt * dt                       'INa, dy/dt * dt
        O_Na = O_Na + dO_Nadt * dt
        I_Na = I_Na + dI_Nadt * dt

        us_CaL = us_CaL + dus_CaLdt * dt
        vC_CaL = vC_CaL + dvC_CaLdt * dt                 'ICaL V-gate, dy/dt * dt
        O_CaL = O_CaL + dO_CaLdt * dt
        I_CaL = I_CaL + dI_CaLdt * dt

        U_CaL = U_CaL + dU_CaLdt * dt
        UCa_CaL = UCa_CaL + dUCa_CaLdt * dt                 'ICaL Ca-gate, dy/dt * dt
        cC_CaL = cC_CaL + dcC_CaLdt * dt

        d_CaT = d_CaT + dd_CaTdt * dt                          'ICaT
        f_CaT = f_CaT + df_CaTdt * dt

        O_K1 = O_K1 + dO_K1dt * dt

        y1_Kr = y1_Kr + dy1_Krdt * dt
        y2_Kr = y2_Kr + dy2_Krdt * dt
        y3_Kr = y3_Kr + dy3_Krdt * dt

        y1_Ks = y1_Ks + dy1_Ksdt * dt
        y2_Ks = y2_Ks + dy2_Ksdt * dt

        y1_Kto = y1_Kto + dy1_Ktodt * dt
        y2_Kto = y2_Kto + dy2_Ktodt * dt

        PXi_NaCa = PXi_NaCa + dPXi_NaCadt * dt
        PXi_NaK = PXi_NaK + dPXi_NaKdt * dt
        PXi_CaP = PXi_CaP + dPXi_CaPdt * dt
        C_RyR = C_RyR + dC_RyRdt * dt                                   'dy/dt * dt
        O_RyR = O_RyR + dO_RyRdt * dt                                   'dy/dt * dt


        totCa_cyt = totCa_cyt + dtotCa_cytdt * dt
        totCa_SRrel = totCa_SRrel + dtotCa_SRreldt * dt
        Ca_SRup = Ca_SRup + dCa_SRupdt * dt                      'dy/dt * dt

        Ki = Ki + dKidt * dt                                              'dy/dt * dt
        Nai = Nai + dNaidt * dt

        TCa = TCa + dTCadt * dt                                 'dy/dt * dt
        TCaCB = TCaCB + dTCaCBdt * dt                  'dy/dt * dt
        TCB = TCB + dTCBdt * dt                               'dy/dt * dt
        h_CB = h_CB + dh_CBdt * dt                           'dy/dt * dt

        ATP = ATP + dATPdt * dt
        If ATP < 0.0001 Then ATP = 0.0001 'it should not be minus.

    End Sub

    Public Sub ATPConcentrations()
        ADP = NucleotideTotal - ATP                                                                  '
        dATPdt = alfaADP * effiADP * ADP - (INaK / F / Vi + IsrCaPup / F / 4 / Vi + 0.4 * TCaCB * TRPNt)               '[ATP]

        ATPbyINaK = INaK / F / Vi * dt                                           'd[ATP] by Na-K pump    stoichiometry 1 e:1ATP   'mM
        ATPbyIsrCaPup = IsrCaPup / F / 4 / Vi * dt                          'd[ATP] by SR Ca pump  stoichiometry 4 e:1ATP   'mM
        ATPbyCntr = 0.4 * TCaCB * TRPNt * dt                                'd[ATP] by Contraction                                         'mM
        SumATPbyPumps = SumATPbyPumps + (ATPbyINaK + ATPbyIsrCaPup)    'Sum of ATP consumption by ion pumps.
        SumATPbyCntr = SumATPbyCntr + ATPbyCntr                                 'Sum of ATP consumption by contraction

        If half_SL < minL Then minL = half_SL 'detect minimum value of halfSL
        If half_SL > maxL Then maxL = half_SL 'detect maximum value of halfSL

    End Sub

    Public Sub Energy()

        CellENGfromSumATPbyCntr = SumATPbyCntr * Vi / 10 ^ 15 * ATPenergy * 10 ^ 12    '(pJ/cell)
        ' concentration(1E-3 mol/L)* (cell volume)(1E-15 L) *  ATP energy(1E3 joul/mol ATP)   '(J/cell)
    End Sub

    '*******************   Contraction  *************************************
    Public TRPNt As Double = 0.07                 'mM, Total concentration of TROPONIN

    Public Tfree As Double             'T 
    Public TCa As Double               'TCa
    Public TCaCB As Double          'TCaCB
    Public TCB As Double               'TCB
    Public dTCadt As Double
    Public dTCaCBdt As Double
    Public dTCBdt As Double
    Public availableFrac As Double                     'Fraction of effective TCa to attach the head.
    Public EffectiveF As Double = 20                'length dependent rate of attaching  '20/um^2

    Public bias As Double = 3                             'adjusting factor for temperature effect
    Public a1_TRP As Double = 39             'Ca binding rate   39 uM/s,  39 1000mM/1000 ms=39 mM/ms
    Public a2_TRP As Double = bias * 0.0013 '* 3 'on-rate of attachment      '1.3/s=1.3/1000ms=0.0013/ms
    Public a3_TRP As Double = 0.05    '0.03           'off-rate of Ca from active cross bridge  '30/s=30/1000ms=0.03/ms
    Public a4_TRP As Double = bias * 0.04 'deactivation rate I  '40/s=40/1000ms=0.04/ms
    Public a5_TRP As Double = bias * 9000 / 2 'deactivation rate II  '9s/um^2,  ??= 9/1000ms/um^2
    Public b1_TRP As Double = 0.05  '0.03         'off rate for Ca from trpn     '30/s=30/1000ms=0.03/ms
    Public b2_TRP As Double = bias * 0.0013 ' * 3 'off rate of attachment    '1.3/s=1.3/1000ms=0.0013/ms
    Public b3_TRP As Double = 1560         'Ca binding rate     '1560 uM/s, 1560 1000 mM/1000 ms=1560
    Public Q1, Q2, Q3, Q4, Q5, Q6 As Double    'net movement of each step    Public half_SL As Double                                 'half sarcomere length         = 1.1     '1.05 um

    Public half_SL As Double      'half sarcomere length 
    Public tempHalf_SL As Double                       'used when new sarcomere length is calculated
    Public limithalf_SL As Double = 1.17        'upper limit of HSL to enable bridge attachment     '1.17 um
    Public x_SL As Double          'half length of thick filament + thin filament length over the non-overlap zone
    Public h_CB As Double         'h
    Public dh_CBdt As Double

    Public ZeroForceEL As Double = 0.97        'Length of parallel elestic component in the absence of load   '0.97 um
    Public limith_CB As Double = 0.005      'lower limit of cross bridge lendgth      '0.005 um
    Public SlidingRate As Double = 1.2            'turn over rate of cross bridge sliding   '1200/s=1200/1000ms=1.2/ms
    Public Resolution As Double                           ' = 0.0001    'resolution in determining new half sarcomere length.

    Public F_Ec As Double                      'Force of the parallel elastic element
    Public Kf_Ec As Double = 140000    'stiffness of the parallel elastic element of a cell    '140000mN/mm^2/um^5
    Public KfL_Ec As Double = 200     'stiffness of the parallel elastic element of a cell, linear component
    Public F_CB As Double                            'force of the myofilament component
    Public Af_CB As Double = 1800000       'Amplitude factor for cross-bridge force of a cell   '1800mN/mm^2/um/uM=1800000 mN/mm^2/um/mM
    Public F_Ext As Double                           'Load applied to the muscle(mN/mm^2)
    Public netF_new As Double                           'used when new sarcomere length is calculated
    Public netF_old As Double                          'used when new sarcomere length is calculated
    Public avlN_CB As Double                           'used when new sarcomere length is calculated

    Public CaTropC As Double                           'net released ca from troponinC
    Public maxF_CB As Double

    Public Sub Contraction()

        Tfree = 1 - TCa - TCaCB - TCB                                                                 'total of TRPN = 1
        availableFrac = 1 / Math.Exp(EffectiveF * (half_SL - limithalf_SL) ^ 2)
        Q1 = a1_TRP * Caf_cyt * Tfree - b1_TRP * TCa                                      ' T  <=====> TCa
        Q2 = a2_TRP * TCa * availableFrac - b2_TRP * TCaCB                         ' TCa <=====> TCa*
        Q3 = a3_TRP * TCaCB - b3_TRP * Caf_cyt * TCB                                   ' TCa* <======> T*
        Q4 = a4_TRP * TCB + a5_TRP * dh_CBdt ^ 2 * TCB                              'T* ---------> T
        Q5 = a5_TRP * dh_CBdt ^ 2 * TCaCB                                                      'TCa* ---------> T

        dTCadt = Q1 - Q2                                                                                  '
        dTCaCBdt = Q2 - Q3 - Q5                                ' - Q6
        dTCBdt = Q3 - Q4

        '(B) ----  Calculate the rate of change in the cross bridge length, dh/dt --------------------------

        avlN_CB = Af_CB * TRPNt * (TCaCB + TCB)     ' available number of CB
        F_CB = avlN_CB * h_CB                                      'Fb = A * (T + TCA) * h                    '
        x_SL = half_SL - h_CB                                         'X = L - h
        tempHalf_SL = half_SL                                        'initial value to determine the new half sarcomere length.

        F_Ec = Kf_Ec * (ZeroForceEL - tempHalf_SL) ^ 5 + KfL_Ec * (ZeroForceEL - tempHalf_SL)  'Fp; parallel elastic component

        netF_new = (F_Ext + F_Ec - F_CB)
        If netF_new > 0 Then Resolution = 0.000005 'elongation of cross bridge
        If netF_new < 0 Then Resolution = -0.000005 'shortening of cross bridge
        Do                                              'determine L (new half sarcomere length)
            netF_old = netF_new
            If netF_old = 0 Then Exit Do
            h_CB = h_CB + Resolution     'change in h
            tempHalf_SL = x_SL + h_CB       'change in halfSL
            If tempHalf_SL > limithalf_SL Then Exit Do 'avoid endless cycle due to an error.
            F_Ec = Kf_Ec * (ZeroForceEL - tempHalf_SL) ^ 5 + KfL_Ec * (ZeroForceEL - tempHalf_SL)
            F_CB = avlN_CB * h_CB
            netF_new = (F_Ext + F_Ec - F_CB)
        Loop While (netF_new * netF_old) > 0

        dh_CBdt = -SlidingRate * (h_CB - limith_CB)                 'dh/dt
        half_SL = tempHalf_SL                                                     'new half_SL

        CaTropC = TRPNt * (Q3 + Q5 - Q1)                 'd(Ca released from TRPN)/dt

    End Sub

    Public Sub externalIon()
        ' ************* Calculate  gradual changes of external ion concentrations and EGTA concentration within the cell.
        Cae = Cae + BathDiffusion * dt * (newCae - Cae)            'simulate exchange of bath solution
        Nae = Nae + BathDiffusion * dt * (newNae - Nae)           'simulate exchange of bath solution
        Ke = Ke + BathDiffusion * dt * (newKe - Ke)                   'simulate exchange of bath solution

        'ACh = ACh + BathDiffusion * dt * (newACh - ACh)

    End Sub


End Class
