﻿'Aim. This model is developed to demonstrate mechanisms underlying  the pacing atrial cells by the natural pacemaker
' cells in the sinoatrial node. 
'80 Atrial cells are connected to 80 sinoatrial node cells. Cell models are based on Kyoto model published in Jap J Physiology.
'The gap junction conductance was adjusted to obtain the conduction velocity of a few cm per sec in the SA node, and to 
'the velocity of ~60 cm/s in the atrial tissue as demonstrated in the graph of cellular conduction velocity.
'The dependency of the conduction velocity can be demonstrated by giving different gap junction conductance in the text Boxes.
'When identical SA node cell models were connected, those cells that are close to Atrial cell array shows marked
' activation of the hyperpolarization activated current, as can be seen in the picture box of membrane current.
' The time course of activation is quite similar to the gap junction current
' to drive the atrial cells at the border between Atrial and SA node cell array.
'SA node cells located at the free end of SA nodel cell array takes the role of pacemaker. The amplitude of the hyperpolarization-activated
' current is quite small, because of low membrane potential.

'The measurements of conduction velocity is valid only when two sequential action potentials are generated within a single sweep.

'Protocol
'Questions


'Cell model is in SACellmodel.vb and ACellmodel.vb
'Experimental procedure is described in ProtocolSAAC.vb.
'Graphic routine is in Graph2014.vb.
'Disk I/O is prepared in calculate.vb.
'Initial values are written and read in CSVRW.vb (csv read & write).
'This model is developed and coded by A. Noma and Iwataki in February 2015.

Imports System.IO
Imports System.Math


Public Class ProtocolSAAC
    '****************  The method of defining many cells using the memory array was introduced by Iwai  in October 2012. 
    '***************** Graph *********************************
    Public RedPen As Pen = New Pen(Color.Red, 1)          'define variable pens in this project 
    Public LimePen As Pen = New Pen(Color.Lime, 1)
    Public BluePen As Pen = New Pen(Color.Blue, 1)
    Public GreenPen As Pen = New Pen(Color.Green, 1)
    Public BlackPen As Pen = New Pen(Color.Black, 1)
    Public GreenYellowPen As Pen = New Pen(Color.GreenYellow, 1)
    Public LightCoralPen As Pen = New Pen(Color.LightCoral, 1)
    Public ChocolatePen As Pen = New Pen(Color.Chocolate, 1)
    Public ChocolatePen3 As Pen = New Pen(Color.Chocolate, 3)
    Public MagentaPen As Pen = New Pen(Color.Magenta, 1)
    Public OlivePen As Pen = New Pen(Color.Olive, 1)
    Public VioletPen As Pen = New Pen(Color.Violet, 1)
    Public GrayPen As Pen = New Pen(Color.Gray, 1)
    Public DarkCyanPen As Pen = New Pen(Color.DarkCyan, 1)
    Public DarkMagentaPen As Pen = New Pen(Color.DarkMagenta, 1)
    Public OrangePen As Pen = New Pen(Color.Orange, 1)

    Public BlackPen2 As Pen = New Pen(Color.Black, 2)
    Public BlackPen4 As Pen = New Pen(Color.Black, 4)
    Public RedPen4 As Pen = New Pen(Color.Red, 4)
    Public cyanPen As Pen = New Pen(Color.Cyan, 1)
    Public yellowPen As Pen = New Pen(Color.Yellow, 1)
    'Public gradient As New Graph.gradientPen
    Public ChocolatePen4 As Pen = New Pen(Color.Chocolate, 4)

    '***************** Graph2014 *********************************
    Public RedBrush As SolidBrush = New SolidBrush(Color.Red)
    Public BlueBrush As SolidBrush = New SolidBrush(Color.Blue)
    Public BlackBrush As SolidBrush = New SolidBrush(Color.Black)
    Public GreenBrush As SolidBrush = New SolidBrush(Color.Green)
    Public ChocolateBrush As SolidBrush = New SolidBrush(Color.Green)

    Public gradient As New Graph2014.GradientPen(3)

    '*****************************************  controling Graphics ***********************************
    Dim graphCurrent As Graph2014                 'This graph shows current of ion channels
    Dim graphVm As Graph2014                        'This graph shows membrane potential of the cell
    Dim graphArray As Graph2014                     'illustrate the cell model in a array
    Dim graphConc As Graph2014
    Dim graphImSA1 As Graph2014
    Dim graphImACNc As Graph2014
    Dim graphVelocity As Graph2014

    Dim MinX1 As Double = 0                            'X-axis parameters, minimum value
    Dim MaxX1 As Double = 800                       'maximum value on the X-axis
    Dim IntervalX1 As Double = 100                 'interval of the grids

    Dim MinY1 As Double = -600                        ' graphCurrent
    Dim MaxY1 As Double = 600
    Dim IntervalY1 As Double = 100

    Dim MinY2 As Double = -90                         'graphVm
    Dim MaxY2 As Double = 50
    Dim IntervalY2 As Double = 10

    Dim MinX3 As Double = 0
    Dim MaxX3 As Double = 160
    Dim IntervalX3 As Double = 1

    Dim MinY3 As Double = 0                         'graph at the bottom = graphVm
    Dim MaxY3 As Double = 2
    Dim IntervalY3 As Double = 1

    Dim MinY4 As Double = 0                         'Concentrations of cations = graph[C]
    Dim MaxY4 As Double = 15
    Dim IntervalY4 As Double = 1

    Dim MinX4 As Double = 0                            'X-axis parameters, minimum value
    Dim MaxX4 As Double = 160                          'maximum value on the X-axis = number of cells
    Dim IntervalX4 As Double = 10                      'interval of the grids

    Dim MinX5 As Double = MinX1                            'X-axis parameters, minimum value
    Dim MaxX5 As Double = MaxX1                          'maximum value on the X-axis = number of cells
    Dim IntervalX5 As Double = IntervalX1 * 2              'interval of the grids

    Dim MinY5 As Double = -150
    Dim MaxY5 As Double = 150
    Dim intervalY5 As Double = 50

    Dim MinY6 As Double = -5000
    Dim MaxY6 As Double = 5000
    Dim intervalY6 As Double = 1000

    Dim MinX7 As Double = 0                            'X-axis parameters, minimum value
    Dim MaxX7 As Double = 160                          'maximum value on the X-axis = number of cells
    Dim IntervalX7 As Double = 10                      'interval of the grids

    Dim MinY7 As Double = 0
    Dim MaxY7 As Double = 100                            'cm/sec    conduction velocity
    Dim intervalY7 As Double = 10


    '*****************************************  controling the Experimental time ***********************************
    Dim dt As Double = 0.03       'time step for integration  by Euler Method
    Dim MaxRepeat = 1                     'number of repetitive sweeps
    Public sweepTime As Double                       'time for a single sweep
    Dim TtoPlot As Double                             'time to plot
    Dim maxTP As Double = 0.5                    'msec, define plotting time interval 

    Dim ontime As Double                              'time to start current pulse
    Dim offtime As Double                             'time to switch off the current pulse
    Dim Amplitude1 As Double                      'pA, amplitude of the pipette current to the left end
    Dim Amplitude2 As Double                      'pA, amplitude of the pipette current to the right end

    Public Nc_SA As Integer = 80                   'number of cell models
    Public SA(80) As SACellmodel                 'define SA cells
    Public Nc_AC As Integer = 80                   'number of cell models
    Public AC(80) As ACellmodel                    'define AC cells

    Dim FileNameSA As String = "APSANcellModel.CSV"
    Dim FileNameAC As String = "APACcellModel.CSV"

    Dim namesSA() As String = New String() {"Vm", "us_Na", "C_Na", "O_Na", "I_Na", "us_CaL", "O_CaL", "vC_CaL", _
             "I_CaL", "U_CaL", "UCa_CaL", "cC_CaL", "d_CaT", "f_CaT", "d_st", "f_st", "us_st", "C1_ha", "C2_ha", "O1_ha", "O2_ha", "O_ACh", _
             "O_K1", "y1_Kr", "y2_Kr",  "y3_Kr", "y1_Ks", "y2_Ks", "y1_Kto", "y2_Kto", "PXi_NaCa", "PXi_NaK", _
             "PXi_CaP", "C_RyR", "O_RyR", "TRPNf", "TCa", "TCaCB", "TCB", "h_CB", _
             "half_SL", "Ke", "Nae", "Cae", "ATP", "Ki", "Nai", "totCa_cyt", "totCa_SRrel", "Ca_SRup"}
    Dim namesAC() As String = New String() {"Vm", "us_Na", "C_Na", "O_Na", "I_Na", "us_CaL", "O_CaL", "vC_CaL", _
             "I_CaL", "U_CaL", "UCa_CaL", "cC_CaL", "d_CaT", "f_CaT", "O_K1", "y1_Kr", "y2_Kr", "y3_Kr", "y1_Ks", "y2_Ks", _
             "y1_Kto", "y2_Kto", "PXi_NaCa", "PXi_NaK", "PXi_CaP", "C_RyR", "O_RyR", "TRPNf", "TCa", "TCaCB", "TCB", _
            "h_CB", "half_SL", "Ke", "Nae", "Cae", "ATP", "Ki", "Nai", "totCa_cyt", "totCa_SRrel", "Ca_SRup"}

    Dim ChartTime As Double = 0                'sweep time to be recorded  

    Public cellTime(200) As Double                    'time at Vm crossing -20 mV
    Public Nsaac As Long
    Public velocity(200) As Double                       'um/msec    conduction velocity between coupled two cells
    Dim cellLength As Double
    Public SAlength As Double = 40                    'um,  SA cell length
    Public AClength As Double = 120                 'um,  AC cell length
    Public ACGgap As Double                              'nS, AC cell temporal Ggap
    Public SAGgap As Double                              'nS, SA cell temporal Ggap
    Public FirstDefinition As Boolean = "true"

    Private Sub cbStart_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles cbStart.Click
        cbStart.Visible = False                     'hide the Start button      

        If FirstDefinition Then                        'avoiding repetive definitions
            defineGraph()                                        'create instants of graph
            Clear()                                                     'once clear all graph
            instanceCell()                                        'create instance of cells
            '************ Read CSV file and convert variables ************************
            CSVRW.ReadCSV(FileNameSA, SA, namesSA)
            CSVRW.ReadCSV(FileNameAC, AC, namesAC)
        End If

        ReadForm()                                          ' read all the values on the Form
        Nsaac = Nc_SA + Nc_AC                    'sequential numbering of all cells of SA and AC

        '************** clear the memory for time_20 detection **********
        For i As Integer = 1 To Nc_SA
            SA(i).detected = False
            SA(i).time_over20 = 0
        Next i
        For i As Integer = 1 To Nc_AC
            AC(i).detected = False
            AC(i).time_over20 = 0
        Next i
        '*************************************
        Do                                                                 'External 'Do loop' for sequential sweeps
            DrawAxes()                                              'clear and rewrite the grids　　グラフ枠を書き換える
            SA(Nc_SA).sumQ = 0
            sweepTime = 0                                        'reset sweepTime　　　　　　　時間を初期設定する　　　　
            Do While sweepTime < MaxX1
                My.Application.DoEvents()                  'detect any events

                Calculate()                                              'Calculate the models , electrical currents and the membrane potential　　　　　　　　　　　　　　

                If TtoPlot >= maxTP Then
                    plotResults()                                        'plot results of calculation  
                    DrawCell()                                            'Draw rectangle cell in gradient pen
                    TtoPlot = 0                                            'reset the counter
                Else
                    TtoPlot = TtoPlot + dt
                End If

                sweepTime = sweepTime + dt          'proceed with the experimental time
                ChartTime = ChartTime + dt           'proceed with the recording interval
            Loop                                                          ' END of one sweep                                   

            calVelocity()                                            'calculate conduction velocity for each cell
            For i As Integer = 2 To Nsaac
                graphVelocity.PlotFillPoint(i - 1, velocity(i), RedBrush, 6)                       'Atrial cell
            Next i

            txtSAV.Text = Format(velocity(40), "##.## cm/s")
            txtACV.Text = Format(velocity(120), "##.## cm/s")


            If RBWrite.Checked Then
                CSVRW.CreateNewCsv(FileNameSA, 1, SA, 1, namesSA)   'Replace the old CSV file with a new one, and prepare the headings
                CSVRW.WriteCsv(FileNameSA, sweepTime, SA, 1, namesSA)   'input = SA, output = CsvFileNameSA 

                CSVRW.CreateNewCsv(FileNameAC, 1, AC, 1, namesAC)   'Replace the old CSV file with a new one, and prepare the headings
                CSVRW.WriteCsv(FileNameAC, sweepTime, AC, 1, namesAC)   'input = AC, output = CsvFileNameAC 
            End If

            txtNai.Text = Format(SA(Nc_SA).Nai, "##.###")
            txtKi.Text = Format(SA(Nc_SA).Ki, "###.###")
            txtsumQ.Text = Format(SA(Nc_SA).sumQ, "####.#")
        Loop While ChartTime <= MaxX1 * MaxRepeat

        cbStart.Visible = True                     'display the button  
    End Sub

    Private Sub calVelocity()                                            'calculate conduction velocity for each cell
        For i As Integer = 1 To Nc_SA                      'determine the time to pass Vm = -20 mV for each SA_cell
            SA(i).time_20 = (SA(i).time_over20 - dt) + dt * (-20 - SA(i).previousVm) / (SA(i).latestVm - SA(i).previousVm)
            cellTime(i) = SA(i).time_20                           'put the velocity into the memory array cellint(J)
        Next

        For i As Integer = 1 To Nc_AC                      'determine the time to pass Vm = -20 mV for each AC_cell
            AC(i).time_20 = (AC(i).time_over20 - dt) + dt * (-20 - AC(i).previousVm) / (AC(i).latestVm - AC(i).previousVm)
            cellTime(Nc_SA + i) = AC(i).time_20                         'put the velocity into the memory array cellint(J)
        Next

        Dim J As Long
        cellLength = SAlength                                  'default case
        For J = 2 To Nsaac
            If J > Nc_SA Then cellLength = AClength 'replace the cell length by the AC length
            velocity(J) = 0.1 * cellLength / (cellTime(J) - cellTime(J - 1))
        Next


    End Sub


    Private Sub WriteCSV_SA()

        Dim Filename1 As String
        Dim sw1 As StreamWriter
        Filename1 = Path.Combine(Application.StartupPath, FileNameSA)
        sw1 = New StreamWriter(Filename1, False)
        ' ***** prepare appropriate number of memory array, and a list of variables to be written in the memory array
        For i As Integer = 1 To Nc_SA       'repeat for all cell numbers,  
            sw1.Write("{0:E}" & "," & "{1:E}" & "," & "{2:E}" & "," & "{3:E}" & "," & "{4:E}" & "," & "{5:E}" & "," & "{6:E}" & "," _
           & "{7:E}" & "," & "{8:E}" & "," & "{9:E}" & "," & "{10:E}" & "," & "{11:E}" & "," & "{12:E}" & "," & "{13:E}" & "," _
           & "{14:E}" & "," & "{15:E}" & "," & "{16:E}" & "," & "{17:E}" & "," & "{18:E}" & "," & "{19:E}" & "," & "{20:E}" & "," _
           & "{21:E}" & "," & "{22:E}" & "," & "{23:E}" & "," & "{24:E}" & "," & "{25:E}" & "," & "{26:E}" & "," & "{27:E}" & "," _
           & "{28:E}" & "," & "{29:E}" & "," & "{30:E}" & "," & "{31:E}" & "," & "{32:E}" & "," & "{33:E}" & "," & "{34:E}" & "," _
           & "{35:E}" & "," & "{36:E}" & "," & "{37:E}" & "," & "{38:E}" & "," & "{39:E}" & "," & "{40:E}" & "," & "{41:E}" & "," _
           & "{42:E}" & "," & "{43:E}" & "," & "{44:E}" & "," & "{45:E}" & "," & "{46:E}" & "," & "{47:E}" & "," & "{48:E}" & "," _
           & "{49:E}" & "," & "{50:E}" & ",", _
             SA(i), SA(i).Vm, SA(i).us_Na, SA(i).C_Na, SA(i).O_Na, SA(i).I_Na, SA(i).us_CaL, SA(i).O_CaL, SA(i).vC_CaL, _
             SA(i).I_CaL, SA(i).U_CaL, SA(i).UCa_CaL, SA(i).cC_CaL, SA(i).d_CaT, SA(i).f_CaT, SA(i).d_st, SA(i).f_st, _
             SA(i).us_st, SA(i).C1_ha, SA(i).C2_ha, SA(i).O1_ha, SA(i).O2_ha, SA(i).O_K1, SA(i).y1_Kr, SA(i).y2_Kr, _
             SA(i).y3_Kr, SA(i).y1_Ks, SA(i).y2_Ks, SA(i).y1_Kto, SA(i).y2_Kto, SA(i).O_ACh, SA(i).PXi_NaCa, SA(i).PXi_NaK, _
             SA(i).PXi_CaP, SA(i).C_RyR, SA(i).O_RyR, SA(i).TCa, SA(i).TCaCB, SA(i).TCB, SA(i).h_CB, _
             SA(i).half_SL, SA(i).Ke, SA(i).Nae, SA(i).Cae, SA(i).ATP, SA(i).Ki, SA(i).Nai, SA(i).totCa_cyt, SA(i).totCa_SRrel, _
             SA(i).Caf_SRrel, SA(i).Ca_SRup)

            sw1.Write(Chr(13) & Chr(10))
        Next

        sw1.Close()
    End Sub

    Private Sub WriteCSV_AC()

        Dim Filename1 As String
        Dim sw1 As StreamWriter
        Filename1 = Path.Combine(Application.StartupPath, FileNameAC)
        sw1 = New StreamWriter(Filename1, False)
        ' ***** prepare appropriate number of memory array, and a list of variables to be written in the memory array
        For i As Integer = 1 To Nc_AC       'repeat for all cell numbers,  
            sw1.Write("{0:E}" & "," & "{1:E}" & "," & "{2:E}" & "," & "{3:E}" & "," & "{4:E}" & "," & "{5:E}" & "," & "{6:E}" & "," _
           & "{7:E}" & "," & "{8:E}" & "," & "{9:E}" & "," & "{10:E}" & "," & "{11:E}" & "," & "{12:E}" & "," & "{13:E}" & "," _
           & "{14:E}" & "," & "{15:E}" & "," & "{16:E}" & "," & "{17:E}" & "," & "{18:E}" & "," & "{19:E}" & "," & "{20:E}" & "," _
           & "{21:E}" & "," & "{22:E}" & "," & "{23:E}" & "," & "{24:E}" & "," & "{25:E}" & "," & "{26:E}" & "," & "{27:E}" & "," _
           & "{28:E}" & "," & "{29:E}" & "," & "{30:E}" & "," & "{31:E}" & "," & "{32:E}" & "," & "{33:E}" & "," & "{34:E}" & "," _
           & "{35:E}" & "," & "{36:E}" & "," & "{37:E}" & "," & "{38:E}" & "," & "{39:E}" & "," & "{40:E}" & "," & "{41:E}" & "," _
           & "{42:E}" & ",", _
             AC(i), AC(i).Vm, AC(i).us_Na, AC(i).C_Na, AC(i).O_Na, AC(i).I_Na, AC(i).us_CaL, AC(i).O_CaL, AC(i).vC_CaL, _
             AC(i).I_CaL, AC(i).U_CaL, AC(i).UCa_CaL, AC(i).cC_CaL, AC(i).d_CaT, AC(i).f_CaT, AC(i).O_K1, AC(i).y1_Kr, AC(i).y2_Kr, _
             AC(i).y3_Kr, AC(i).y1_Ks, AC(i).y2_Ks, AC(i).y1_Kto, AC(i).y2_Kto, AC(i).PXi_NaCa, AC(i).PXi_NaK, _
             AC(i).PXi_CaP, AC(i).C_RyR, AC(i).O_RyR, AC(i).TRPNf, AC(i).TCa, AC(i).TCaCB, AC(i).TCB, AC(i).h_CB, _
             AC(i).half_SL, AC(i).Ke, AC(i).Nae, AC(i).Cae, AC(i).ATP, AC(i).Ki, AC(i).Nai, AC(i).totCa_cyt, AC(i).totCa_SRrel, _
              AC(i).Ca_SRup)

            sw1.Write(Chr(13) & Chr(10))
        Next

        sw1.Close()
    End Sub

    Private Sub ReadCSV_SA()

        'StreamReader.ReadLine メソッド を使ってCSVファイルを読み込む()
        Dim Filename1 As String
        Filename1 = Path.Combine(Application.StartupPath, FileNameSA)   'path of the project
        Dim sr1 As StreamReader

        Try
            sr1 = New StreamReader(Filename1)

            For i As Integer = 1 To Nc_SA                 'repeat for all Cell numbers
                Dim temp(50) As String                 'prepare a memory array 0 ~ 50
                temp = Split(sr1.ReadLine(), ",")    'fill the memory array "temp"

                SA(i).Vm = CDbl(temp(1))                'get value from the memory array address 1

                SA(i).us_Na = CDbl(temp(2))            'get value from the memory array address 2
                SA(i).C_Na = CDbl(temp(3))
                SA(i).O_Na = CDbl(temp(4))
                SA(i).I_Na = CDbl(temp(5))

                SA(i).us_CaL = CDbl(temp(6))
                SA(i).O_CaL = CDbl(temp(7))
                SA(i).vC_CaL = CDbl(temp(8))
                SA(i).I_CaL = CDbl(temp(9))
                SA(i).U_CaL = CDbl(temp(10))
                SA(i).UCa_CaL = CDbl(temp(11))
                SA(i).cC_CaL = CDbl(temp(12))

                SA(i).d_CaT = CDbl(temp(13))
                SA(i).f_CaT = CDbl(temp(14))

                SA(i).d_st = CDbl(temp(15))
                SA(i).f_st = CDbl(temp(16))
                SA(i).us_st = CDbl(temp(17))

                SA(i).C1_ha = CDbl(temp(18))
                SA(i).C2_ha = CDbl(temp(19))
                SA(i).O1_ha = CDbl(temp(20))
                SA(i).O2_ha = CDbl(temp(21))

                SA(i).O_K1 = CDbl(temp(22))

                SA(i).y1_Kr = CDbl(temp(23))
                SA(i).y2_Kr = CDbl(temp(24))
                SA(i).y3_Kr = CDbl(temp(25))

                SA(i).y1_Ks = CDbl(temp(26))
                SA(i).y2_Ks = CDbl(temp(27))

                SA(i).y1_Kto = CDbl(temp(28))
                SA(i).y2_Kto = CDbl(temp(29))

                SA(i).O_ACh = CDbl(temp(30))

                SA(i).PXi_NaCa = CDbl(temp(31))
                SA(i).PXi_NaK = CDbl(temp(32))
                SA(i).PXi_CaP = CDbl(temp(33))

                SA(i).C_RyR = CDbl(temp(34))
                SA(i).O_RyR = CDbl(temp(35))

                SA(i).TCa = CDbl(temp(36))
                SA(i).TCaCB = CDbl(temp(37))
                SA(i).TCB = CDbl(temp(38))
                SA(i).h_CB = CDbl(temp(39))
                SA(i).half_SL = CDbl(temp(40))

                SA(i).Ke = CDbl(temp(41))
                SA(i).Nae = CDbl(temp(42))
                SA(i).Cae = CDbl(temp(43))

                SA(i).ATP = CDbl(temp(44))
                SA(i).Ki = CDbl(temp(45))
                SA(i).Nai = CDbl(temp(46))
                SA(i).totCa_cyt = CDbl(temp(47))

                SA(i).totCa_SRrel = CDbl(temp(48))
                SA(i).Caf_SRrel = CDbl(temp(49))
                SA(i).Ca_SRup = CDbl(temp(49))
            Next

        Catch ex As Exception
            If TypeOf (ex) Is FileNotFoundException Then
                MessageBox.Show("ファイル?が見つかりまUせんでした?。B")
            ElseIf TypeOf (ex) Is InvalidCastException OrElse TypeOf (ex) Is NullReferenceException Then
                MessageBox.Show("ファイルの形式が不正です。B")
            Else
                MessageBox.Show(ex.Message)
            End If

            For i As Integer = 1 To Nc_SA
                SA(i).Vm = -59.9712112942936                    '0, membrane potential   

                SA(i).us_Na = 0.0344642287551903             '1, INa ultra slow gate
                SA(i).C_Na = 0.00226121207995763            '2, Na channel closed state
                SA(i).O_Na = 0.0000395261155015815        '3, Na channel open state
                SA(i).I_Na = 0.921986586043834                 '4, Na channel inactivated state

                SA(i).us_CaL = 0.813870603250832              '5, s    ICa,L
                SA(i).O_CaL = 0.000605101949876967          '6. ICaL shirokov pAP
                SA(i).vC_CaL = 0.437601116758376              '7, ICaL shirokov pRP
                SA(i).I_CaL = 0.55778166367865                   '8, ICaL shirokov pAI
                SA(i).U_CaL = 0.0495890125282116              '9, ICaL shirokov pU
                SA(i).UCa_CaL = 0.000135271748430735      '10, ICaL shirokov pUCa
                SA(i).cC_CaL = 0.127312941219539               '11, ICaL shirokov pC

                SA(i).d_CaT = 0.00190466924478167              '12, d    ICa,T
                SA(i).f_CaT = 0.0456352016531766                 '13, f    ICa,T

                SA(i).d_st = 0.13376395831656                        '14, d    Ist
                SA(i).f_st = 0.319108656567219                       '15, f     Ist
                SA(i).us_st = 0.542986646350026                     '16, s    Ist  slow gate

                SA(i).C1_ha = 0.999448037573957                    '17, C1    iha
                SA(i).C2_ha = 0.000508077691621167              '18, C2     iha
                SA(i).O1_ha = 0.000041290576217986              '19, O1    iha
                SA(i).O2_ha = 0.00000247175679383289          '20, O2    iha

                SA(i).O_K1 = 0.397363677276551                     '21, iK1

                SA(i).y1_Kr = 0.578033804322097                     '22,  af   ikr
                SA(i).y2_Kr = 0.644740367142839                     '23, as   ikr
                SA(i).y3_Kr = 0.859735717402154                     '24, i    ikr

                SA(i).y1_Ks = 0.271780917923823                     '25, i    iks
                SA(i).y2_Ks = 0.176272995137349                     '26, i    iks

                SA(i).y1_Kto = 0.00299803019642708                '27, ito
                SA(i).y2_Kto = 0.855189237449591                    '28, ito

                SA(i).O_ACh = 0.0198352531955566                  '29, iACh

                SA(i).PXi_NaCa = 0.955983768483093               '30, INaCa  PXI
                SA(i).PXi_NaK = 0.644975156076689                 '31, INaK  PXI
                SA(i).PXi_CaP = 0.839741889009139                  '32, SRCaPPXI       Vm = mery(0)

                SA(i).C_RyR = 0.0073583281314326                    '33, RyR
                SA(i).O_RyR = 0.0035292160881915                    '34, RyR

                SA(i).TCa = 0.132237169504377                           '35, Troponin-FCa
                SA(i).TCaCB = 0.0207680241009334                    '36, TroponinCa-CB
                SA(i).TCB = 0.00313730913363329                      '37, Troponin-CB
                SA(i).h_CB = 0.00506381981151422                     '38, Cross Bridge Length
                SA(i).half_SL = 0.962379990000417                     '39, half sarcomere.

                SA(i).Ke = 5.4                                                      '40, External K   concentration,   [K]e
                SA(i).Nae = 140                                                   '41, External Na concentration,   [Na]e
                SA(i).Cae = 1.8                                                    '42, External Ca concentration,   [Ca]e

                SA(i).ATP = 4.86338546118028                           '43, intracelluler ATP concentratiion. (mM)

                SA(i).Ki = 142.251803316812                              '44, [K]i
                SA(i).Nai = 4.95140579421772                            '45, [Na]i
                SA(i).totCa_cyt = 0.0027556004027566               '46, [Ca]i      total Ca

                SA(i).totCa_SRrel = 6.76638847879939               '47, [Ca]rel   total  Ca in the releasing site
                SA(i).Caf_SRrel = 1.06203870659193                  '48, [Ca]rel   free Ca
                SA(i).Ca_SRup = 3.46318127053819                    '49, [Ca]up
            Next i

        Finally               '例外が発生してもしなても必ず実行する
            'If Not sr1 Is Nothing Then      'FileのI読み込みに成功した
            '    sr1.Close()
            'End If
        End Try

    End Sub

    Private Sub ReadCSV_AC()
        '********************* Atrial cell ***********************************************
        'StreamReader.ReadLine メソッド を使ってCSVファイルを読み込む()
        Dim Filename2 As String
        Filename2 = Path.Combine(Application.StartupPath, FileNameAC)   'path of the project

        Dim sr2 As StreamReader = Nothing      'キーワード　何も参照されていないことを示す

        Try       '例外処理の実行　実行時エラーの可能性がある場合とりあえずトライする
            sr2 = New StreamReader(Filename2)     'ストリームリーダを参照し、ファイルからデータを獲得する

            For i As Integer = 1 To Nc_AC                 'repeat for all Cell numbers
                Dim temp(50) As String                 'prepare a memory array 0 ~ 50
                temp = Split(sr2.ReadLine(), ",")    'fill the memory array "temp"

                AC(i).Vm = CDbl(temp(1))                'get value from the memory array address 1

                AC(i).us_Na = CDbl(temp(2))            'get value from the memory array address 2
                AC(i).C_Na = CDbl(temp(3))
                AC(i).O_Na = CDbl(temp(4))
                AC(i).I_Na = CDbl(temp(5))

                AC(i).us_CaL = CDbl(temp(6))
                AC(i).O_CaL = CDbl(temp(7))
                AC(i).vC_CaL = CDbl(temp(8))
                AC(i).I_CaL = CDbl(temp(9))
                AC(i).U_CaL = CDbl(temp(10))
                AC(i).UCa_CaL = CDbl(temp(11))
                AC(i).cC_CaL = CDbl(temp(12))

                AC(i).d_CaT = CDbl(temp(13))
                AC(i).f_CaT = CDbl(temp(14))

                AC(i).O_K1 = CDbl(temp(15))

                AC(i).y1_Kr = CDbl(temp(16))
                AC(i).y2_Kr = CDbl(temp(17))
                AC(i).y3_Kr = CDbl(temp(18))

                AC(i).y1_Ks = CDbl(temp(19))
                AC(i).y2_Ks = CDbl(temp(20))

                AC(i).y1_Kto = CDbl(temp(21))
                AC(i).y2_Kto = CDbl(temp(22))

                AC(i).PXi_NaCa = CDbl(temp(23))
                AC(i).PXi_NaK = CDbl(temp(24))
                AC(i).PXi_CaP = CDbl(temp(25))

                AC(i).C_RyR = CDbl(temp(26))
                AC(i).O_RyR = CDbl(temp(27))

                AC(i).TRPNf = CDbl(temp(28))

                AC(i).TCa = CDbl(temp(29))
                AC(i).TCaCB = CDbl(temp(30))
                AC(i).TCB = CDbl(temp(31))
                AC(i).h_CB = CDbl(temp(32))
                AC(i).half_SL = CDbl(temp(33))

                AC(i).Ke = CDbl(temp(34))
                AC(i).Nae = CDbl(temp(35))
                AC(i).Cae = CDbl(temp(36))

                AC(i).ATP = CDbl(temp(37))
                AC(i).Ki = CDbl(temp(38))
                AC(i).Nai = CDbl(temp(39))
                AC(i).totCa_cyt = CDbl(temp(40))

                'AC(i).totCa_SRrel = CDbl(temp(48))
                AC(i).Caf_SRrel = CDbl(temp(41))
                AC(i).Ca_SRup = CDbl(temp(42))
            Next

        Catch ex As Exception        'ex 例外クラスの定義　　ファイルの読み込みができなかった場合
            If TypeOf (ex) Is FileNotFoundException Then
                MessageBox.Show("ファイルが見つかりませんでした。")
            ElseIf TypeOf (ex) Is InvalidCastException OrElse TypeOf (ex) Is NullReferenceException Then
                MessageBox.Show("ファイルの形式が不正です。")
            Else
                MessageBox.Show(ex.Message)
            End If

            '********  fileの読み込みに失敗したので、以下のデフォルト値を入力する  *********
            For i As Integer = 1 To Nc_AC
                AC(i).Vm = -82.8          'membrane potential  

                AC(i).us_Na = 0.5667451              'INa ultra slow gate
                AC(i).C_Na = 0.69559625                 'Na channel closed state
                AC(i).O_Na = 0.000071269              'Na channel open state
                AC(i).I_Na = 0.0878997               'Na channel inactivated state

                AC(i).us_CaL = 0.999873411                          's    ICa,L
                AC(i).O_CaL = 0.00000777                          'ICaL shirokov pAP
                AC(i).vC_CaL = 0.9890205                          'ICaL shirokov pRP
                AC(i).I_CaL = 0.0001553964                          'ICaL shirokov pAI
                AC(i).U_CaL = 0.037235517                         'ICaL shirokov pU
                AC(i).UCa_CaL = 0.0005704985                     'ICaL shirokov pUCa
                AC(i).cC_CaL = 0.372368287                           'ICaL shirokov pC

                AC(i).d_CaT = 0.0000159744                          'd    ICa,T
                AC(i).f_CaT = 0.86422383                       'f    ICa,T

                AC(i).O_K1 = 0.61422141                                      'iK1

                AC(i).y1_Kr = 0.0019106                     'af   ikr
                AC(i).y2_Kr = 0.2113425                       'as   ikr
                AC(i).y3_Kr = 0.96846802                         'i    ikr

                AC(i).y1_Ks = 0.108276985                              'i    iks
                AC(i).y2_Ks = 0.02769158                        'i    iks

                AC(i).y1_Kto = 0.00078473                          'ito
                AC(i).y2_Kto = 0.999999147                               'ito

                AC(i).PXi_NaCa = 0.992149405                   'INaCa  PXI

                AC(i).PXi_NaK = 0.6055062                    'INaK  PXI

                AC(i).PXi_CaP = 0.4185431                  'SRCaPPXI       Vm = mery(0)

                AC(i).C_RyR = 0.228638476                    'RyR
                AC(i).O_RyR = 0.00022071                   'RyR

                AC(i).TRPNf = 0.0064                                    'T, troponin
                AC(i).TCa = 0.01833810672                             'TCa,  Troponin-FCa
                AC(i).TCaCB = 0.001437596                        'TCA*,    TroponinCa-CB
                AC(i).TCB = 0.00032507                        'T*,    Troponin-CB
                AC(i).h_CB = 0.004995192                        'Cross Bridge Length
                AC(i).half_SL = 0.96442                  'half sarcomere.

                '******************* substrates ****************
                AC(i).Ke = 5.4                      'External K   concentration,   [K]e
                AC(i).Nae = 140                       'External Na concentration,   [Na]e
                AC(i).Cae = 1.8                       'External Ca concentration,   [Ca]e

                AC(i).ATP = 5.002                      'intracelluler ATP concentratiion. (mM)

                AC(i).Ki = 142.19                                   '[K]i
                AC(i).Nai = 5.26                              '[Na]i
                AC(i).totCa_cyt = 0.000278                           '[Ca]i      total Ca
                AC(i).totCa_SRrel = 9.1325                         '[Ca]rel   total  Ca in the releasing site

                AC(i).Ca_SRup = 2.178                        '[Ca]SRup  
            Next i

        Finally               '例外が発生してもしなくても必ず実行する
            If Not sr2 Is Nothing Then      'Fileの読み込みに成功した。
                sr2.Close()
            End If
        End Try

    End Sub

    Private Sub instanceCell()
        For i As Integer = 0 To Nc_SA        'The array number should start from 0
            SA(i) = New SACellmodel
        Next

        For i As Integer = 0 To Nc_AC        'The array number should start from 0
            AC(i) = New ACellmodel
        Next

    End Sub

    Private Sub Calculate()

        '*************************************** IgapK ********************
        For i As Integer = 1 To Nc_SA - 1
            SA(i).Igapjunc(SA(i).IgapK, SA(i).Ggap, SA(i).Vm, SA(i + 1).Vm, SA(i).Ki, SA(i + 1).Ki)  ' 
        Next
        For i As Integer = 1 To Nc_AC - 1
            AC(i).Igapjunc(AC(i).IgapK, AC(i).Ggap, AC(i).Vm, AC(i + 1).Vm, AC(i).Ki, AC(i + 1).Ki)  '
        Next

        SA(Nc_SA).Igapjunc(SA(Nc_SA).IgapK, SA(Nc_SA).Ggap, SA(Nc_SA).Vm, AC(1).Vm, SA(Nc_SA).Ki, AC(1).Ki)
        AC(Nc_AC).IgapK = 0

        SA(Nc_SA).sumQ = SA(Nc_SA).sumQ + SA(Nc_SA).IgapK * dt / AC(1).F                   'mmol, sum of IgapK between SA(Nc) and AC(1)

        '*************************************** IextK ********************
        SA(1).IextK = SA(1).IgapK                  'left edge

        For i As Integer = 2 To Nc_SA
            SA(i).Iexternal(SA(i).IextK, SA(i - 1).IgapK, SA(i).IgapK)      'sum the two gap junctional currents on both sides for the current integration
        Next

        AC(1).IextK = -SA(Nc_SA).IgapK + AC(1).IgapK          'connection between SA and A cell arrays

        For i As Integer = 2 To Nc_AC
            AC(i).Iexternal(AC(i).IextK, AC(i - 1).IgapK, AC(i).IgapK)      'sum the two gap junctional currents on both sides for the current integration
        Next

        For i As Integer = 1 To Nc_SA
            SA(i).Preparedydt(dt)
            SA(i).EulerMethod(dt)
            If SA(i).detected = False Then
                If SA(i).Vm > -20 Then    'Vm passed -20 mV level for the first time and time-20 is still not measured.
                    SA(i).time_over20 = sweepTime                               'time_over20 is determined
                    SA(i).latestVm = SA(i).Vm                                      'presentVm is determined
                    SA(i).detected = True
                Else                                                                                'Vm is still more negative than -20 mV,
                    SA(i).previousVm = SA(i).Vm                                    'replace previousVm by the new Vm
                End If
            Else
            End If
        Next

        For i As Integer = 1 To Nc_AC
            AC(i).Preparedydt(dt)
            AC(i).EulerMethod(dt)
            If AC(i).detected = False Then
                If AC(i).Vm > -20 Then    'Vm passed -20 mV level for the first time and time-20 is still not measured.
                    AC(i).time_over20 = sweepTime                               'time_over20 is determined
                    AC(i).latestVm = AC(i).Vm                                     'presentVm is determined   
                    AC(i).detected = True
                Else                                                                              'Vm is still more negative than -20 mV,
                    AC(i).previousVm = AC(i).Vm                                    'replace previousVm by the new Vm
                End If
            Else
            End If

        Next

    End Sub

    Private Sub plotResults()

        graphVm.PlotFillPoint(sweepTime, SA(1).Vm, BlackBrush, 2)
        graphVm.PlotFillPoint(sweepTime, SA(10).Vm, BlackBrush, 2)
        graphVm.PlotFillPoint(sweepTime, SA(20).Vm, BlackBrush, 2)
        graphVm.PlotFillPoint(sweepTime, SA(30).Vm, BlackBrush, 2)
        graphVm.PlotFillPoint(sweepTime, SA(40).Vm, BlackBrush, 2)
        graphVm.PlotFillPoint(sweepTime, SA(50).Vm, BlackBrush, 2)
        graphVm.PlotFillPoint(sweepTime, SA(60).Vm, BlackBrush, 2)
        graphVm.PlotFillPoint(sweepTime, SA(70).Vm, BlackBrush, 2) 'plot results
        graphVm.PlotFillPoint(sweepTime, SA(75).Vm, BlackBrush, 2)
        graphVm.PlotFillPoint(sweepTime, SA(76).Vm, BlackBrush, 2)
        graphVm.PlotFillPoint(sweepTime, SA(77).Vm, BlackBrush, 2)
        graphVm.PlotFillPoint(sweepTime, SA(78).Vm, BlackBrush, 2)
        graphVm.PlotFillPoint(sweepTime, SA(79).Vm, BlackBrush, 2)
        graphVm.PlotFillPoint(sweepTime, SA(80).Vm, BlackBrush, 2) 'plot results

        graphVm.PlotFillPoint(sweepTime, AC(1).Vm, RedBrush, 2)
        graphVm.PlotFillPoint(sweepTime, AC(2).Vm, RedBrush, 2)
        graphVm.PlotFillPoint(sweepTime, AC(3).Vm, RedBrush, 2)
        graphVm.PlotFillPoint(sweepTime, AC(4).Vm, RedBrush, 2)
        graphVm.PlotFillPoint(sweepTime, AC(5).Vm, RedBrush, 2)
        graphVm.PlotFillPoint(sweepTime, AC(10).Vm, RedBrush, 2) 'plot results
        graphVm.PlotFillPoint(sweepTime, AC(20).Vm, RedBrush, 2)
        graphVm.PlotFillPoint(sweepTime, AC(30).Vm, RedBrush, 2)
        graphVm.PlotFillPoint(sweepTime, AC(40).Vm, RedBrush, 2)
        graphVm.PlotFillPoint(sweepTime, AC(50).Vm, RedBrush, 2)
        graphVm.PlotFillPoint(sweepTime, AC(60).Vm, RedBrush, 2)
        graphVm.PlotFillPoint(sweepTime, AC(70).Vm, RedBrush, 2) 'plot results
        graphVm.PlotFillPoint(sweepTime, AC(80).Vm, RedBrush, 2)

        graphCurrent.PlotFillPoint(sweepTime, SA(Nc_SA).IgapK, BlackBrush, 3)
        'graphCurrent.PlotFillPoint(sweepTime, SA(Nc_SA).Iha, chocolateBrush, 3)
        Dim sumIha As Double = 0
        graphCurrent.PlotFillPoint(sweepTime, SA(Nc_SA).Ibnsc, RedBrush, 3)
        sumIha = sumIha + SA(80).Iha
        graphCurrent.PlotFillPoint(sweepTime, sumIha, ChocolateBrush, 3)
        sumIha = sumIha + SA(79).Iha
        graphCurrent.PlotFillPoint(sweepTime, sumIha, ChocolateBrush, 3)
        sumIha = sumIha + SA(78).Iha
        graphCurrent.PlotFillPoint(sweepTime, sumIha, ChocolateBrush, 3)
        sumIha = sumIha + SA(77).Iha
        graphCurrent.PlotFillPoint(sweepTime, sumIha, ChocolateBrush, 3)
        sumIha = sumIha + SA(76).Iha
        graphCurrent.PlotFillPoint(sweepTime, sumIha, ChocolateBrush, 3)
        sumIha = sumIha + SA(75).Iha
        graphCurrent.PlotFillPoint(sweepTime, sumIha, ChocolateBrush, 3)

        'graphCurrent.PlotFillPoint(sweepTime, SA(Nc_SA - 1).Iha, RedBrush, 3)
        'graphCurrent.PlotFillPoint(sweepTime, (SA(Nc_SA).Iha + SA(Nc_SA - 1).Iha + SA(Nc_SA - 2).Iha), RedBrush, 3)

        graphImSA1.PlotFillPoint(sweepTime, SA(Nc_SA).IKr, GreenBrush, 3)
        graphImSA1.PlotFillPoint(sweepTime, SA(Nc_SA).ICaL, BlackBrush, 3)
        'graphImSA1.PlotFillPoint(sweepTime, SA(Nc_SA).Ist, GreenBrush, 2)
        graphImSA1.PlotFillPoint(sweepTime, SA(1).Iha, RedBrush, 3)                   'SA(1).Iha
        'graphImSA1.PlotFillPoint(sweepTime, SA(Nc_SA - 20).Iha, RedBrush, 3)
        'graphImSA1.PlotFillPoint(sweepTime, SA(Nc_SA - 10).Iha, RedBrush, 3)
        'graphImSA1.PlotFillPoint(sweepTime, SA(Nc_SA).Iha, RedBrush, 3)

        graphImACNc.PlotFillPoint(sweepTime, AC(1).ICaL * AC(Nc_AC).Cm, BlackBrush, 3)
        graphImACNc.PlotFillPoint(sweepTime, AC(1).IKr * AC(Nc_AC).Cm, GreenBrush, 3)
        graphImACNc.PlotFillPoint(sweepTime, AC(1).IKpl * AC(Nc_AC).Cm, RedBrush, 3)
        graphImACNc.PlotFillPoint(sweepTime, AC(1).IK1 * AC(Nc_AC).Cm, BlueBrush, 3)

    End Sub

    Private Sub ReadForm()                         ' read all the values on the Form

        MaxRepeat = CDbl(txtMaxRepeat.Text)          'Number of sweeps 

        ACGgap = CDbl(txtGgapA.Text)                        'nS, Ggap of AC
        SAGgap = CDbl(txtGgapSA.Text)                       'nS, Ggap of SA

        For i As Integer = 1 To Nc_SA                           'define all of SA node cells
            SA(i).Cm = SA(i).originalCm                          'membrane capacitance Cm
            SA(i).Ggap = SAGgap                                   'gap junction conductance on the right side of individual cells
            SA(i).sfbNSC = 0.1                                      'scaling factor for the non-selective cation current 
        Next i
        If chkDisconnect.Checked Then SA(Nc_SA).Ggap = 0 'If checked, then SA(last number).Ggap is nullified

        '********** leading pacemaker cells are defined by scaling factor of IbNSC *******
        SA(1).sfbNSC = 1
        SA(2).sfbNSC = 1
        SA(3).sfbNSC = 1
        SA(4).sfbNSC = 1
        SA(5).sfbNSC = 1
        SA(6).sfbNSC = 1
        SA(7).sfbNSC = 1
        SA(8).sfbNSC = 1
        SA(9).sfbNSC = 1
        SA(10).sfbNSC = 1
        SA(11).sfbNSC = 1
        SA(12).sfbNSC = 1
        SA(13).sfbNSC = 1
        SA(14).sfbNSC = 1
        SA(15).sfbNSC = 1
        SA(16).sfbNSC = 0.8
        SA(17).sfbNSC = 0.6
        SA(18).sfbNSC = 0.4
        SA(19).sfbNSC = 0.2

        For i As Integer = 1 To Nc_AC
            AC(i).Cm = AC(i).originalCm
            AC(i).Ggap = ACGgap
        Next
    End Sub

    Public Sub calculateIinject(ByVal time As Double, ByVal ontime As Double, ByVal offtime As Double, ByVal Amp As Double, ByRef Iinject As Double)
        If time >= ontime And time <= offtime Then
            Iinject = Amp
        Else
            Iinject = 0
        End If
    End Sub

    Private Sub defineGraph()                                        ' create instances of Graph
        graphCurrent = New Graph2014(PBcurrent, MinX1, MaxX1, MinY1, MaxY1)
        graphVm = New Graph2014(PBpotential, MinX1, MaxX1, MinY2, MaxY2)
        graphArray = New Graph2014(pbArray, MinX3, MaxX3, MinY3, MaxY3)
        graphImSA1 = New Graph2014(pbIm1, MinX5, MaxX5, MinY5, MaxY5)
        graphImACNc = New Graph2014(pbImNc, MinX5, MaxX5, MinY6, MaxY6)
        graphVelocity = New Graph2014(pbVelocity, MinX7, MaxX7, MinY7, MaxY7)
    End Sub

    Private Sub DrawAxes()                                           ' draw axis
        graphCurrent.DrawAxis(IntervalX1, IntervalY1, " ", 8)
        graphVm.DrawAxis(IntervalX1, IntervalY2, " ", 8)
        graphArray.DrawAxis(IntervalX3, IntervalY3, " ", 8)
        graphImSA1.DrawAxis(IntervalX5, intervalY5, "", 8)                             'draw grid of graph
        graphImACNc.DrawAxis(IntervalX5, intervalY6, "", 8)                             'draw grid of graph
        graphVelocity.DrawAxis(IntervalX7, intervalY7, "", 8)                             'draw grid of graph

        graphArray.PlotLine(Nc_SA, MinY3, Nc_SA, MaxY3, ChocolatePen4)

    End Sub

    Private Sub Clear()                                           ' draw axis
        graphCurrent.Clear()
        graphVm.Clear()
        graphArray.Clear()
        graphImSA1.Clear()                             'draw grid of graph
        graphImACNc.Clear()                             'draw grid of graph
        graphVelocity.Clear()                             'draw grid of graph

        graphArray.PlotLine(Nc_SA, MinY3, Nc_SA, MaxY3, ChocolatePen4)

    End Sub

    Private Sub DrawCell()
        For i As Integer = 1 To Nc_SA
            graphArray.PlotRectangle(gradient.PenGreenBlue(SA(i).Vm, -80, 10), i - 1, 1, 0.7, 1)
        Next i

        For i As Integer = 1 To Nc_AC
            graphArray.PlotRectangle(gradient.PenGreenBlue(AC(i).Vm, -80, 10), Nc_SA + i - 1, 1, 0.7, 1)
        Next i


    End Sub



    Private Sub Button1_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Button1.Click
        sweepTime = MaxX1 + 1
        ChartTime = (MaxRepeat + 1) * MaxX1
    End Sub

    Private Sub cbErase_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles cbErase.Click
        graphCurrent.Clear()
        graphVm.Clear()
        graphArray.Clear()

        DrawAxes()
    End Sub

    Private Sub BEnd_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles BEnd.Click
        End
    End Sub

    Private Sub Label6_Click(sender As System.Object, e As System.EventArgs)

    End Sub
End Class
